"""
openinverter custom SDO protocol unit tests
Utility structure tests
"""

import unittest

from openinverter_can_tool.oi_node import CanMessage, MapEntry

# Reduce test verbosity
# pylint: disable=missing-function-docstring


class TestCanMessages(unittest.TestCase):
    """Verify the creation of CanMessage instances"""

    def test_can_message_with_negative_can_id_fails(self):
        with self.assertRaises(ValueError):
            CanMessage(-1, [], False)

    def test_can_message_with_too_big_standard_can_id_fails(self):
        with self.assertRaises(ValueError):
            CanMessage(0x800, [], False)

    def test_can_message_with_too_big_extended_can_id_fails(self):
        with self.assertRaises(ValueError):
            CanMessage(0x20000000, [], True)

    def test_can_message_with_zero_can_id_is_allowed(self):
        msg = CanMessage(0, [], False)
        assert msg.can_id == 0
        assert not msg.is_extended_frame

    def test_can_message_with_extended_can_id_is_allowed(self):
        msg = CanMessage(0x12345678, [], True)
        assert msg.can_id == 0x12345678
        assert msg.is_extended_frame


class TestMapEntry(unittest.TestCase):
    """Verify the creation of MapEntry instances"""

    def test_map_entry_with_negative_position_fails(self):
        with self.assertRaises(ValueError):
            MapEntry(
                param_id=1,
                position=-1,
                length=1,
                gain=1,
                offset=0
            )

    def test_map_entry_with_position_beyond_frame_length_fails(self):
        with self.assertRaises(ValueError):
            MapEntry(
                param_id=1,
                position=64,
                length=1,
                gain=1,
                offset=0
            )

    def test_map_entry_with_zero_length_fails(self):
        with self.assertRaises(ValueError):
            MapEntry(
                param_id=1,
                position=0,
                length=0,
                gain=1,
                offset=0
            )

    def test_map_entry_with_length_larger_than_32_bit_word_fails(self):
        with self.assertRaises(ValueError):
            MapEntry(
                param_id=1,
                position=0,
                length=33,
                gain=1,
                offset=0
            )

    def test_map_entry_with_excessive_positive_gain_fails(self):
        with self.assertRaises(ValueError):
            MapEntry(
                param_id=1,
                position=0,
                length=1,
                gain=10000.0,
                offset=0
            )

    def test_map_entry_with_excessive_negative_gain_fails(self):
        with self.assertRaises(ValueError):
            MapEntry(
                param_id=1,
                position=0,
                length=1,
                gain=-10000.0,
                offset=0
            )

    def test_map_entry_with_excessive_positive_offset_fails(self):
        with self.assertRaises(ValueError):
            MapEntry(
                param_id=1,
                position=0,
                length=1,
                gain=1,
                offset=128
            )

    def test_map_entry_with_excessive_negative_offset_fails(self):
        with self.assertRaises(ValueError):
            MapEntry(
                param_id=1,
                position=0,
                length=1,
                gain=1,
                offset=-129
            )


if __name__ == "__main__":
    unittest.main()



"""
Test CAN message map persistence
"""
import filecmp
import unittest
from pathlib import Path

import canopen.objectdictionary
import pytest

from openinverter_can_tool.map_persistence import (export_json_map,
                                                   import_json_map)
from openinverter_can_tool.oi_node import CanMessage, MapEntry
from openinverter_can_tool.paramdb import import_database

MAP_DIR = Path(__file__).parent / "test_data" / "maps"
DB_DIR = Path(__file__).parent / "test_data" / "paramdb"

# Reduce test verbosity
# pylint: disable=missing-function-docstring


class TestJSONMaps:
    """Test CAN message map persistence to and from JSON"""

    def test_export_empty_map(self, tmp_path: Path):
        map_file_path = tmp_path / "empty.json"

        with open(map_file_path, "wt", encoding="utf-8") as map_file:
            export_json_map([], [], canopen.ObjectDictionary(), map_file)

        assert filecmp.cmp(map_file_path, MAP_DIR / "empty.json",
                           shallow=False)

    def test_export_single_tx_message_single_param(self, tmp_path: Path):

        tx_map = [
            CanMessage(0x123, [MapEntry(1, 24, 8, -1.0, 0)])
        ]

        db = import_database(DB_DIR / "single-param.json")

        map_file_path = tmp_path / "single-tx-message-single-param.json"
        with open(map_file_path, "wt", encoding="utf-8") as map_file:
            export_json_map(tx_map, [], db, map_file)

        assert filecmp.cmp(
            map_file_path,
            MAP_DIR / "single-tx-message-single-param.json",
            shallow=False)

    def test_export_simple_tx_and_rx_message_map(self, tmp_path: Path):

        tx_map = [
            CanMessage(0x123, [MapEntry(1, 24, 8, -1.0, 0)]),
        ]
        rx_map = [
            CanMessage(0x321, [MapEntry(1, 23, -16, 2.5, -42)])
        ]

        db = import_database(DB_DIR / "single-param.json")

        map_file_path = tmp_path / "simple-tx-rx-message-map.json"
        with open(map_file_path, "wt", encoding="utf-8") as map_file:
            export_json_map(tx_map, rx_map, db, map_file)

        assert filecmp.cmp(
            map_file_path,
            MAP_DIR / "simple-tx-rx-message-map.json",
            shallow=False)

    def test_export_multiple_tx_messages_with_multiple_params(
            self,
            tmp_path: Path):

        tx_map = [
            CanMessage(0x101, [
                MapEntry(17, 24, 8, -1.0, 0),
                MapEntry(18, 0, 8, 1.0, 0),
                MapEntry(17, 8, 8, -1.0, 0),
                MapEntry(18, 16, 8, 1.0, 0)
            ]),
            CanMessage(0x333, [
                MapEntry(2035, 0, 8, 1.0, 0),
                MapEntry(107, 8, 8, -1.0, 0),
                MapEntry(2035, 16, 8,  1.0, 0),
                MapEntry(107, 24, 8, -1.0, 0)
            ])
        ]

        db = import_database(DB_DIR / "complex.json")

        map_file_path = tmp_path / "multiple-tx-messages.json"
        with open(map_file_path, "wt", encoding="utf-8") as map_file:
            export_json_map(tx_map, [], db, map_file)

        assert filecmp.cmp(
            map_file_path,
            MAP_DIR / "multiple-tx-messages.json",
            shallow=False)

    def test_export_map_with_invalid_param_id(self, tmp_path: Path):

        rx_map = [
            CanMessage(0x123, [MapEntry(2, 24, 8, -1.0, 0)])
        ]

        db = import_database(DB_DIR / "single-param.json")

        map_file_path = tmp_path / "empty-file.json"
        with pytest.raises(KeyError):
            with open(map_file_path, "wt", encoding="utf-8") as map_file:
                export_json_map([], rx_map, db, map_file)

        assert map_file_path.stat().st_size == 0

    def test_import_empty_maps(self):
        db = import_database(DB_DIR / "single-param.json")

        with open(MAP_DIR / "empty.json", "rt", encoding="utf-8") as map_file:
            (tx_map, rx_map) = import_json_map(map_file, db)
            assert not tx_map
            assert not rx_map

    def test_import_single_tx_message_single_param(self):
        db = import_database(DB_DIR / "single-param.json")

        expected_tx_map = [
            CanMessage(0x123, [MapEntry(1, 24, 8, -1.0, 0)])
        ]

        with open(MAP_DIR / "single-tx-message-single-param.json",
                  "rt",
                  encoding="utf-8") as map_file:
            (tx_map, rx_map) = import_json_map(map_file, db)
            assert tx_map == expected_tx_map
            assert not rx_map

    def test_import_simple_tx_and_rx_message_map(self):
        db = import_database(DB_DIR / "single-param.json")

        expected_tx_map = [
            CanMessage(0x123, [MapEntry(1, 24, 8, -1.0, 0)])
        ]
        expected_rx_map = [
            CanMessage(0x321, [MapEntry(1, 23, -16, 2.5, -42)])
        ]

        with open(MAP_DIR / "simple-tx-rx-message-map.json",
                  "rt",
                  encoding="utf-8") as map_file:
            (tx_map, rx_map) = import_json_map(map_file, db)
            assert tx_map == expected_tx_map
            assert rx_map == expected_rx_map

    def test_import_multiple_tx_messages_with_multiple_params(self):
        db = import_database(DB_DIR / "complex.json")

        expected_tx_map = [
            CanMessage(0x101, [
                MapEntry(17, 24, 8, -1.0, 0),
                MapEntry(18, 0, 8, 1.0, 0),
                MapEntry(17, 8, 8, -1.0, 0),
                MapEntry(18, 16, 8, 1.0, 0)
            ]),
            CanMessage(0x333, [
                MapEntry(2035, 0, 8, 1.0, 0),
                MapEntry(107, 8, 8, -1.0, 0),
                MapEntry(2035, 16, 8, 1.0, 0),
                MapEntry(107, 24, 8, -1.0, 0)
            ])
        ]

        with open(MAP_DIR / "multiple-tx-messages.json",
                  "rt",
                  encoding="utf-8") as map_file:
            (tx_map, rx_map) = import_json_map(map_file, db)
            assert repr(tx_map) == repr(expected_tx_map)
            assert not rx_map

    def test_import_unsupported_future_version(self):
        with pytest.raises(RuntimeError, match="Unsupported version: 999"):
            with open(MAP_DIR / "unsupported-version.json",
                      "rt",
                      encoding="utf-8") as map_file:
                import_json_map(
                    map_file, canopen.ObjectDictionary())

    def test_import_invalid_file_format(self):
        with pytest.raises(RuntimeError, match="Invalid file format"):
            with open(DB_DIR / "empty-but-valid.json",
                      "rt",
                      encoding="utf-8") as map_file:
                import_json_map(map_file, canopen.ObjectDictionary())

    def test_import_corrupt_missing_can_id(self):
        db = import_database(DB_DIR / "single-param.json")
        with pytest.raises(KeyError):
            with open(MAP_DIR / "corrupt-missing-can-id.json",
                      "rt",
                      encoding="utf-8") as map_file:
                import_json_map(map_file, db)

    def test_import_corrupt_invalid_param_name(self):
        db = import_database(DB_DIR / "single-param.json")
        with pytest.raises(KeyError):
            with open(MAP_DIR / "corrupt-invalid-param-name.json",
                      "rt",
                      encoding="utf-8") as map_file:
                import_json_map(map_file, db)

    def test_round_trip_complex_tx_and_rx_map(self, tmp_path: Path):

        tx_map = [
            # Simple 8-bit mapping of common temperature values
            CanMessage(0x101, [
                MapEntry(2019, 0, 8, 1.0, 0),
                MapEntry(2020, 8, 8, 1.0, 0)
            ]),

            # 8-bit mapping with an offset
            CanMessage(0x102, [
                MapEntry(2019, 0, 8, 1.0, 10),
                MapEntry(2020, 8, 8, 1.0, 10)
            ]),

            # Little-endian full 32-bit mapping with generous scale
            CanMessage(0x103, [
                MapEntry(2019, 0, 32, 100.0, 0),
                MapEntry(2020, 32, 32, 100.0, 0)
            ]),

            # Big-endian 16-bit mapping of an inverter parameter
            CanMessage(0x104, [
                MapEntry(22, 23, -16, 1.0, 0)
            ]),

            # Big-endian full 32-bit mapping with generous scale
            CanMessage(0x105, [
                MapEntry(2019, 31, -32, 100.0, 0),
                MapEntry(2020, 63, -32, 100.0, 0)
            ])
        ]

        rx_map = [
            # Receive an inverter parameter with a big-endian mapping
            CanMessage(0x201, [
                MapEntry(22, 23, -16, 1.0, 0)
            ])
        ]

        db = import_database(DB_DIR / "mapable-params.json")

        map_file_path = tmp_path / "complex_tx_and_rx_map.json"
        with open(map_file_path, "wt", encoding="utf-8") as map_file:
            export_json_map(tx_map, rx_map, db, map_file)

        with open(map_file_path, "rt", encoding="utf-8") as read_map_file:
            (in_tx_map, in_rx_map) = import_json_map(read_map_file, db)
            assert in_tx_map == tx_map
            assert in_rx_map == rx_map

    def test_import_simple_tx_and_rx_message_map_without_extended_support(
            self):
        """The original file format from 0.0.9"""
        db = import_database(DB_DIR / "single-param.json")

        expected_tx_map = [
            CanMessage(0x123, [MapEntry(1, 24, 8, -1.0, 0)])
        ]
        expected_rx_map = [
            CanMessage(0x321, [MapEntry(1, 23, -16, 2.5, -42)])
        ]

        with open(MAP_DIR /
                  "simple-tx-rx-message-map-without-extended-support.json",
                  "rt",
                  encoding="utf-8") as map_file:
            (tx_map, rx_map) = import_json_map(map_file, db)
            assert tx_map == expected_tx_map
            assert rx_map == expected_rx_map

    def test_import_multiple_tx_messages_with_extended_can_ids(self):
        db = import_database(DB_DIR / "complex.json")

        expected_tx_map = [
            CanMessage(0x101, [
                MapEntry(17, 24, 8, -1.0, 0),
                MapEntry(18, 0, 8, 1.0, 0),
                MapEntry(17, 8, 8, -1.0, 0),
                MapEntry(18, 16, 8, 1.0, 0)
            ], is_extended_frame=True),
            CanMessage(0x12345678, [
                MapEntry(2035, 0, 8, 1.0, 0),
                MapEntry(107, 8, 8, -1.0, 0),
                MapEntry(2035, 16, 8, 1.0, 0),
                MapEntry(107, 24, 8, -1.0, 0)
            ], is_extended_frame=True)
        ]

        with open(MAP_DIR / "multiple-tx-extended-id-messages.json",
                  "rt",
                  encoding="utf-8") as map_file:
            (tx_map, rx_map) = import_json_map(map_file, db)
            assert repr(tx_map) == repr(expected_tx_map)
            assert not rx_map

    def test_import_corrupt_out_of_range_extended_can_id(self):
        db = import_database(DB_DIR / "single-param.json")
        with pytest.raises(ValueError):
            with open(MAP_DIR / "corrupt-out-of-range-extended-can-id.json",
                      "rt",
                      encoding="utf-8") as map_file:
                import_json_map(map_file, db)


if __name__ == '__main__':
    unittest.main()



"""
Test CAN message map to dbc persistence
"""
import unittest
from pathlib import Path

import canopen.objectdictionary
import pytest
from approvaltests import verify

from openinverter_can_tool.map_persistence import (export_dbc_map,
                                                   export_json_map,
                                                   transform_map_to_canopen_db)
from openinverter_can_tool.oi_node import CanMessage, MapEntry
from openinverter_can_tool.paramdb import import_database

DB_DIR = Path(__file__).parent / "test_data" / "paramdb"

# Reduce test verbosity
# pylint: disable=missing-function-docstring


class TestDBCMaps:
    """Test CAN message map persistence to DBC message definition files"""

    def test_transform_empty_map(self):
        canopen_db = transform_map_to_canopen_db(
            None,
            [],
            [],
            canopen.ObjectDictionary())

        verify(canopen_db)

    def test_transform_single_tx_message_single_param(self):

        tx_map = [
            CanMessage(0x123, [MapEntry(1, 24, 8, -1.0, 0)])
        ]

        db = import_database(DB_DIR / "single-param.json")

        canopen_db = transform_map_to_canopen_db(None, tx_map, [], db)

        verify(canopen_db)

    def test_transform_simple_tx_and_rx_message_map(self):

        tx_map = [
            CanMessage(0x123, [MapEntry(1, 24, 8, -1.0, 0)]),
        ]
        rx_map = [
            CanMessage(0x321, [MapEntry(1, 23, -16, 2.5, -42)])
        ]

        db = import_database(DB_DIR / "single-param.json")

        canopen_db = transform_map_to_canopen_db(None, tx_map, rx_map, db)

        verify(canopen_db)

    def test_transform_multiple_tx_messages_with_multiple_params(
            self):

        tx_map = [
            CanMessage(0x101, [
                MapEntry(17, 24, 8, -1.0, 0),
                MapEntry(18, 0, 8, 1.0, 0),
                MapEntry(17, 8, 8, -1.0, 0),
                MapEntry(18, 16, 8, 1.0, 0)
            ]),
            CanMessage(0x333, [
                MapEntry(2035, 0, 8, 1.0, 0),
                MapEntry(107, 8, 8, -1.0, 0),
                MapEntry(2035, 16, 8,  1.0, 0),
                MapEntry(107, 24, 8, -1.0, 0)
            ])
        ]

        db = import_database(DB_DIR / "complex.json")

        canopen_db = transform_map_to_canopen_db(None, tx_map, [], db)

        verify(canopen_db)

    def test_transform_map_with_invalid_param_id(self):

        rx_map = [
            CanMessage(0x123, [MapEntry(2, 24, 8, -1.0, 0)])
        ]

        db = import_database(DB_DIR / "single-param.json")

        with pytest.raises(KeyError):
            transform_map_to_canopen_db(None, [], rx_map, db)

    def test_export_multiple_tx_messages_with_multiple_params(
            self, tmp_path: Path):

        tx_map = [
            CanMessage(0x101, [
                MapEntry(17, 24, 8, -1.0, 0),
                MapEntry(18, 0, 8, 1.0, 0),
                MapEntry(17, 8, 8, -1.0, 0),
                MapEntry(18, 16, 8, 1.0, 0)
            ]),
            CanMessage(0x333, [
                MapEntry(2035, 0, 8, 1.0, 0),
                MapEntry(107, 8, 8, -1.0, 0),
                MapEntry(2035, 16, 8,  1.0, 0),
                MapEntry(107, 24, 8, -1.0, 0)
            ])
        ]

        db = import_database(DB_DIR / "complex.json")

        dbc_path = tmp_path / "multiple_tx_messages_with_multiple_params.dbc"
        export_dbc_map(None, tx_map, [], db, dbc_path)

        with open(dbc_path, "rt", encoding="utf-8") as dbc_file:
            verify(dbc_file.read())

    def test_export_complex_tx_and_rx_map(
            self, tmp_path: Path):

        tx_map = [
            # Simple 8-bit mapping of common temperature values
            CanMessage(0x101, [
                MapEntry(2019, 0, 8, 1.0, 0),
                MapEntry(2020, 8, 8, 1.0, 0)
            ]),

            # 8-bit mapping with an offset
            CanMessage(0x102, [
                MapEntry(2019, 0, 8, 1.0, 10),
                MapEntry(2020, 8, 8, 1.0, 10)
            ]),

            # Little-endian full 32-bit mapping with generous scale
            CanMessage(0x103, [
                MapEntry(2019, 0, 32, 100.0, 0),
                MapEntry(2020, 32, 32, 100.0, 0)
            ]),

            # Big-endian 16-bit mapping of an inverter parameter
            CanMessage(0x104, [
                MapEntry(22, 23, -16, 1.0, 0)
            ]),

            # Big-endian full 32-bit mapping with generous scale
            CanMessage(0x105, [
                MapEntry(2019, 31, -32, 100.0, 0),
                MapEntry(2020, 63, -32, 100.0, 0)
            ])
        ]

        rx_map = [
            # Receive an inverter parameter with a big-endian mapping
            CanMessage(0x201, [
                MapEntry(22, 23, -16, 1.0, 0)
            ])
        ]

        db = import_database(DB_DIR / "mapable-params.json")

        dbc_path = tmp_path / "complex_tx_and_rx_map.dbc"
        export_dbc_map(None, tx_map, rx_map, db, dbc_path)

        with open(dbc_path, "rt", encoding="utf-8") as dbc_file:
            verify(dbc_file.read())

        # export a JSON map to allow the resulting DBC to be manually tested
        # with SavvyCAN on real hardware
        map_file_path = tmp_path / "complex_tx_and_rx_map.json"
        with open(map_file_path, "wt", encoding="utf-8") as map_file:
            export_json_map(tx_map, rx_map, db, map_file)

    def test_export_tx_map_with_enum_param(self, tmp_path: Path):

        tx_map = [
            # Map an enum param to the first 8 bits of a frame
            CanMessage(1, [
                MapEntry(95, 0, 8, 1.0, 0),
            ]),
        ]

        db = import_database(DB_DIR / "mapable-params.json")

        dbc_path = tmp_path / "tx_map_with_enum_param.dbc"
        export_dbc_map(None, tx_map, [], db, dbc_path)

        with open(dbc_path, "rt", encoding="utf-8") as dbc_file:
            verify(dbc_file.read())

    def test_export_tx_map_with_bitfield_spot_value(self, tmp_path: Path):

        tx_map = [
            # Map a bitfield spot value param to the first 8 bits of a frame
            CanMessage(1, [
                MapEntry(2044, 0, 8, 1.0, 0),
            ]),
        ]

        db = import_database(DB_DIR / "mapable-params.json")

        dbc_path = tmp_path / "tx_map_with_bitfield_spot_value.dbc"
        export_dbc_map(None, tx_map, [], db, dbc_path)

        with open(dbc_path, "rt", encoding="utf-8") as dbc_file:
            verify(dbc_file.read())

    def test_export_tx_and_rx_map_with_node_name(self, tmp_path: Path):

        tx_map = [
            CanMessage(1, [MapEntry(2019, 0, 8, 1.0, 0)])
        ]

        rx_map = [
            CanMessage(2, [MapEntry(22, 32, 16, 1.0, 0)])
        ]

        db = import_database(DB_DIR / "mapable-params.json")

        dbc_path = tmp_path / "tx_and_rx_map_with_node_name.dbc"
        export_dbc_map("custom_node_name", tx_map, rx_map, db, dbc_path)

        with open(dbc_path, "rt", encoding="utf-8") as dbc_file:
            verify(dbc_file.read())

    def test_export_tx_message_with_extended_frame_ids(self, tmp_path: Path):

        tx_map = [
            CanMessage(0x00000123, [MapEntry(1, 24, 8, -1.0, 0)], True),
            CanMessage(0x12345678, [MapEntry(1, 24, 8, -1.0, 0)], True)
        ]

        db = import_database(DB_DIR / "single-param.json")

        dbc_path = tmp_path / "tx_message_with_extended_frame_ids.dbc"
        export_dbc_map(None, tx_map, [], db, dbc_path)

        with open(dbc_path, "rt", encoding="utf-8") as dbc_file:
            verify(dbc_file.read())

        # export a JSON map to allow the resulting DBC to be manually tested
        # with SavvyCAN on real hardware
        map_file_path = tmp_path / "tx_message_with_extended_frame_ids.json"
        with open(map_file_path, "wt", encoding="utf-8") as map_file:
            export_json_map(tx_map, [], db, map_file)

    def test_export_tx_map_with_enum_param_with_offset_and_gain(
            self, tmp_path: Path):

        tx_map = [
            # Map an enum param to the first 8 bits of a frame
            # Add an offset and gain to muddle the actual transmitted values
            CanMessage(1, [
                MapEntry(95, 0, 8, 5.0, 13),
            ]),
        ]

        db = import_database(DB_DIR / "mapable-params.json")

        dbc_path = tmp_path / "tx_map_with_enum_param_and_offset.dbc"
        export_dbc_map(None, tx_map, [], db, dbc_path)

        with open(dbc_path, "rt", encoding="utf-8") as dbc_file:
            verify(dbc_file.read())

    def test_export_tx_map_with_bitfield_spot_value_and_offset_and_gain(
            self, tmp_path: Path):

        tx_map = [
            # Map a bitfield spot value param to the first 8 bits of a frame
            # add an offset and gain to muddle the actual transmitted values
            CanMessage(1, [
                MapEntry(2044, 0, 16, 4.0, 1),
            ]),
        ]

        db = import_database(DB_DIR / "mapable-params.json")

        dbc_path = tmp_path / "tx_map_with_bitfield_spot_value_and_offset.dbc"
        export_dbc_map(None, tx_map, [], db, dbc_path)

        with open(dbc_path, "rt", encoding="utf-8") as dbc_file:
            verify(dbc_file.read())


if __name__ == '__main__':
    unittest.main()



"""
openinverter parameter database functions
"""


import json
import re
from pathlib import Path
from typing import Dict, Optional

import canopen
import canopen.objectdictionary

from .fpfloat import fixed_from_float
from .remote_db import RemoteDatabaseNode


def is_power_of_two(num: int) -> bool:
    """Use some clever bitwise anding and arithmetic to determine wether a
    number is a power of two"""
    return (num != 0) and (num & (num - 1) == 0)


def is_bitfield(values: Dict[int, str]) -> bool:
    """Try to figure out if the dictionary of values is a bitfield or an
    enumeration"""

    for value in values:
        # Ignore zero
        if (value != 0) and (not is_power_of_two(value)):
            return False

    # When we only have two non-zero values we assume it's an enum not a
    # bitfield
    if len(values) <= 3:
        return False
    else:
        return True


def filter_zero_bytes(database_bytes: bytes) -> str:
    """Remove any zero bytes before decoding as utf-8. This can be used to
    deal with erroneous data that can be sent by heavily loaded openinverter
    firmware."""
    database_bytes = database_bytes.replace(b"\x00", b"")
    database_str = database_bytes.decode(encoding="utf-8", errors="ignore")
    return database_str


class OIVariable(canopen.objectdictionary.Variable):
    """An openinverter parameter variable has several differences from a
    standard CANopen variable. This class wraps up those differences allowing
    parameters to be specified in an object database."""

    def __init__(self, name: str, param_id: int):
        # assign dummy values to the index to allow creation of the parent
        # variable class
        super().__init__(name, 0, 0)

        # Assign the id to set the index and subindex
        self.id = param_id

        # All openinverter params are 32-bit fixed float values
        # we will convert to float on presentation as required
        # but work with them as integers to keep the canopen
        # library happy
        self.factor = 32
        self.data_type = canopen.objectdictionary.INTEGER32

        self.isparam: bool = False
        self.category: Optional[str] = None

        # This replaces the parent's bit_definitions member
        self.bit_definitions: Dict[int, str] = {}

    @property
    def id(self) -> int:
        """The openinverter parameter identifier for this parameter or spot
        value"""
        return ((self.index & ~0x2100) << 8) + self.subindex

    @id.setter
    def id(self, value: int):
        self.index = 0x2100 | (value >> 8)
        self.subindex = value & 0xFF


def import_database_json(
        paramdb_json: dict) -> canopen.ObjectDictionary:
    """Import an openinverter parameter database JSON.

    :param paramdb_json:
        A dictionary containing an openinverter parameter database

    :returns:
        The Object Dictionary.
    :rtype: canopen.ObjectDictionary
    """

    dictionary = canopen.ObjectDictionary()
    for param_name in paramdb_json:
        param = paramdb_json[param_name]

        # Ignore parameters without unique IDs
        if "id" not in param:
            continue

        var = OIVariable(param_name, int(param["id"]))

        # Common attributes for parameters and values
        var.unit = param["unit"]
        var.isparam = param["isparam"]

        if "category" in param:
            var.category = param["category"]

        # parse units containing enumerations or bitfields
        unit = param["unit"]
        if '=' in unit:
            try:
                # Some parameters like "lasterr" have trailing commas
                unit = unit.rstrip(",")

                values = {
                    int(value): description for value, description in [
                        item.split('=') for item in
                        re.split(r'[,\s]', unit) if item]
                }

                # Infer if this a bitfield or an enumeration and store the
                # description in the appropriate dictionary
                if is_bitfield(values):
                    var.bit_definitions = values
                else:
                    var.value_descriptions = values

            except ValueError:
                # If something bad happens parsing the unit string just bail
                # out and make it a plain numerical value
                unit += " [DB FORMAT ERROR]"

        # Store the unit text for all types of parameters
        var.unit = unit

        # Parameters have additional required attributes
        if var.isparam:
            var.min = fixed_from_float(float(param["minimum"]))
            var.max = fixed_from_float(float(param["maximum"]))
            var.default = fixed_from_float(float(param["default"]))

        dictionary.add_object(var)

    return dictionary


def import_database(paramdb: Path) -> canopen.ObjectDictionary:
    """Import an openinverter parameter database file.

    :param paramdb:
        A path to an openinverter parameter database file

    :returns:
        The Object Dictionary.
    :rtype: canopen.ObjectDictionary
    """

    with open(paramdb, encoding="utf-8") as file:
        doc = json.load(file)

    return import_database_json(doc)


def import_remote_database(
        network: canopen.Network,
        node_id: int) -> canopen.ObjectDictionary:
    """Import an openinverter parameter database from a remote node.

    :param network:
        The configured and started canopen.Network to use to communicate with
        the node.

    :param node_id:
        The openinverter node we wish to obtain the parameter database from.

    :returns:
        The Object Dictionary.
    :rtype: canopen.ObjectDictionary
    """
    node = RemoteDatabaseNode(network, node_id)

    return import_database_json(
        json.loads(filter_zero_bytes(node.param_db())))


def import_cached_database(
        network: canopen.Network,
        node_id: int,
        cache_location: Path
) -> canopen.ObjectDictionary:
    """Import an openinverter parameter database from a remote node and cache
    it for quicker access in future.

    :param network:
        The configured and started canopen.Network to use to communicate with
        the node.

    :param node_id:
        The openinverter node we wish to obtain the parameter database from.

    :param cache_location:
        A directory containing the parameter database cache.

    :returns:
        The Object Dictionary.

    :rtype: canopen.ObjectDictionary
    """

    if not cache_location.exists():
        cache_location.mkdir(parents=True)

    node = RemoteDatabaseNode(network, node_id)

    checksum = node.param_db_checksum()

    cache_file = cache_location / f"{node_id}-{checksum}.json"

    if cache_file.exists():
        dictionary = import_database(cache_file)
    else:
        param_db_str = filter_zero_bytes(node.param_db())

        dictionary = import_database_json(json.loads(param_db_str))

        # Only save the database in the cache when we have successfully
        # imported it
        with open(cache_file, "wt", encoding="utf-8") as param_db_file:
            param_db_file.write(param_db_str)

    return dictionary



"""
openinverter CAN Tools main program
"""

import csv
import datetime
import functools
import glob
import json
import logging
import os
import re
import time
from ast import literal_eval
from pathlib import Path
from typing import List, Optional, Union, cast

import appdirs
import can
import canopen
import canopen.objectdictionary
import click

from . import constants as oi
from .can_upgrade import CanUpgrader, Failure, State, StateUpdate
from .fpfloat import fixed_from_float, fixed_to_float
from .map_persistence import export_dbc_map, export_json_map, import_json_map
from .oi_node import CanMessage, Direction, OpenInverterNode
from .paramdb import OIVariable, import_cached_database, import_database


class CliSettings:
    """Simple class to store the common settings used for all commands"""

    def __init__(
            self,
            database_path: str,
            context: str,
            node_number: int,
            timeout: float,
            debug: bool) -> None:
        self.database_path = database_path
        self.context = context
        self.node_number = node_number
        self.network: Optional[canopen.Network] = None
        self.database = canopen.objectdictionary.ObjectDictionary()
        self.node: Optional[OpenInverterNode] = None
        self.timeout = timeout
        self.debug = debug


pass_cli_settings = click.make_pass_decorator(CliSettings)


def db_action(func):
    """Figure out what parameter database we are to use and allow the wrapped
    function to access it. This decorator should be specified prior to the
    can_action decorator to allow the SDO node to be created.
    """
    @functools.wraps(func)
    def wrapper_db_action(*args, **kwargs):

        # Assume that the first argument exists and is a CliSettings
        cli_settings: CliSettings = args[0]

        if cli_settings.database_path:
            device_db = import_database(Path(cli_settings.database_path))
        else:
            # Fire up the CAN network just to grab the node parameter
            # database from the device
            with canopen.Network() as network:
                network.connect(context=cli_settings.context)
                network.check()

                device_db = import_cached_database(
                    network,
                    cli_settings.node_number,
                    Path(appdirs.user_cache_dir(oi.APPNAME, oi.APPAUTHOR)))

        cli_settings.database = device_db

        # Call the command handler function
        return func(*args, **kwargs)

    return wrapper_db_action


def can_action(func):
    """Establish a CAN connection and allow the wrapped function to access
     the configured CAN SDO node
     """
    @functools.wraps(func)
    def wrapper_can_action(*args, **kwargs):

        # Assume that the first argument exists and is a CliSettings
        cli_settings: CliSettings = args[0]

        # Ensure we always have something to return
        return_value = None

        try:
            # Start with creating a network representing one CAN bus
            with canopen.Network() as network:
                # Connect to the CAN bus
                network.connect(context=cli_settings.context)

                network.check()

                node = OpenInverterNode(
                    network,
                    cli_settings.node_number,
                    cli_settings.database)
                node.sdo.RESPONSE_TIMEOUT = cli_settings.timeout

                # store the network and node objects in the context
                cli_settings.network = network
                cli_settings.node = node

                # Call the command handler function
                return_value = func(*args, **kwargs)

        except canopen.SdoAbortedError as err:
            if err.code == oi.SDO_ABORT_OBJECT_NOT_AVAILABLE:
                click.echo("Command or parameter not supported")
            else:
                click.echo(f"Unexpected SDO Abort: {err}")

        except canopen.SdoCommunicationError as err:
            click.echo(f"SDO communication error: {err}")

        except can.exceptions.CanOperationError as err:
            click.echo(f"CAN error: {err}")

        except OSError as err:
            click.echo(f"OS error: {err}")

        return return_value

    return wrapper_can_action


@click.group()
@click.option("-d", "--database",
              type=click.Path(exists=True, file_okay=True, dir_okay=False),
              help="Override the openinverter JSON parameter database to use")
@click.option("-c", "--context",
              default=None,
              show_default=True,
              type=click.STRING,
              help="Which python-can configuration context to use")
@click.option("-n", "--node",
              default=1,
              show_default=True,
              type=click.INT,
              envvar="OIC_NODE",
              show_envvar=True,
              help="The CAN SDO node ID to communicate with")
@click.option("-t", "--timeout",
              default=1.0,
              show_default=True,
              type=click.FLOAT,
              help="Response timeout in seconds")
@click.option("--debug",
              is_flag=True,
              default=False,
              help="Enable detailed debugging messages")
@click.version_option()
@click.pass_context
def cli(ctx: click.Context,
        database: str,
        context: str,
        node: int,
        timeout: float,
        debug: bool) -> None:
    """openinverter CAN Tool allows querying and setting configuration of
    inverter parameters over a CAN connection"""

    if debug:
        logging.basicConfig(level=logging.DEBUG)

    ctx.obj = CliSettings(database, context, node, timeout, debug)


def print_param_def(item: OIVariable) -> None:
    """Output the definition of a single parameter definition"""
    click.echo(f"{item.name} [{item.unit}]", nl=False)

    if item.isparam:
        assert item.min is not None
        assert item.max is not None
        assert item.default is not None
        click.echo(
            f" - min: {fixed_to_float(item.min):g} "
            f"max: {fixed_to_float(item.max):g} "
            f"default: {fixed_to_float(item.default):g}")
    else:
        click.echo(" - read-only value")


@cli.command()
@click.argument("param", required=True)
@pass_cli_settings
@db_action
def listparam(cli_settings: CliSettings, param: str) -> None:
    """List the definition of PARAM"""

    if param in cli_settings.database.names:
        print_param_def(cli_settings.database.names[param])
    else:
        click.echo(f"Unknown parameter: {param}")


@cli.command()
@pass_cli_settings
@db_action
def listparams(cli_settings: CliSettings) -> None:
    """List all available parameters and values"""
    for item in cli_settings.database.names.values():
        print_param_def(item)


def print_param_value(variable: OIVariable, value: float) -> None:
    """Print out the value of a parameter or outputs the enumeration value or
    bits in a bitfield"""

    click.echo(f"{variable.name:20}: ", nl=False)

    if variable.value_descriptions:
        if value in variable.value_descriptions:
            click.echo(f"{variable.value_descriptions[int(value)]}")
        else:
            click.echo(f"{value:g} (Unknown value)")
    elif variable.bit_definitions:
        value = int(value)
        bit_str = ""
        for bit, description in variable.bit_definitions.items():
            if bit & value:
                bit_str = bit_str + description + ", "
        bit_str = bit_str.removesuffix(", ")

        if len(bit_str) == 0:
            if value in variable.bit_definitions:
                bit_str = variable.bit_definitions[value]
            else:
                bit_str = "0"

        click.echo(f"{bit_str}")
    else:
        click.echo(
            f"{value:g} [{variable.unit}]")


@cli.command()
@pass_cli_settings
@db_action
@can_action
def dumpall(cli_settings: CliSettings) -> None:
    """Dump the values of all available parameters and values"""

    node = cli_settings.node
    for item in cli_settings.database.names.values():
        print_param_value(item, fixed_to_float(node.sdo[item.name].raw))


@cli.command()
@click.argument("param", required=True)
@pass_cli_settings
@db_action
@can_action
def read(cli_settings: CliSettings, param: str) -> None:
    """Read the value of PARAM from the device"""

    if param in cli_settings.database.names:
        node = cli_settings.node
        print_param_value(
            cli_settings.database.names[param],
            fixed_to_float(node.sdo[param].raw))
    else:
        click.echo(f"Unknown parameter: {param}")


@cli.command()
@click.argument("params", required=True, nargs=-1)
@click.argument("out_file", type=click.File("w"))
@click.option("-s", "--step",
              default=1,
              show_default=True,
              type=click.IntRange(1, 3600),
              help="Time to wait before querying for new data in seconds")
@click.option("--timestamp/--no-timestamp",
              show_default=True,
              default=True,
              help="Include a timestamp on each row of the log")
@click.option("--symbolic/--numeric",
              show_default=True,
              default=True,
              help="Convert bit-field and enumerations to symbolic "
              "text values or leave as numbers")
@pass_cli_settings
@db_action
@can_action
def log(cli_settings: CliSettings,
        params: tuple,
        out_file: click.File,
        step: int,
        timestamp: bool,
        symbolic: bool) -> None:
    """
    Log the value of PARAMS from the device periodically in CSV
    format. Multiple parameters may be specified separated by a space.
    OUT_FILE may be a filename or - to output to stdout.

    Special wild card parameter names can be used:

    \b
    ALL    - Logs all parameters and spot values on the device
    PARAMS - Logs all the parameters on the device
    VALUES - Logs all spot values on the device

    The wildcards are case-sensitive.
    """

    query_list: List[OIVariable] = []
    avail_names = cli_settings.database.names

    # special case wildcard parameter names
    if "ALL" in params:
        query_list = list(avail_names.values())

    elif "PARAMS" in params:
        query_list = [p for p in avail_names.values() if p.isparam]

    elif "VALUES" in params:
        query_list = [p for p in avail_names.values() if not p.isparam]

    else:
        # Validate the list of supplied parameters
        for param in params:
            if param in avail_names:
                query_list.append(avail_names[param])
            else:
                click.echo(f"Ignoring unknown parameter: {param}")

    # create a CSV writer to control the output in a format that LibreOffice
    # can open and graph easily
    if timestamp:
        header_list = ["timestamp"]
    else:
        header_list = []
    header_list += [p.name for p in query_list]

    writer = csv.DictWriter(
        out_file, fieldnames=header_list, quoting=csv.QUOTE_ALL)
    writer.writeheader()

    # Loop forever logging
    node = cli_settings.node
    while True:
        row = {}
        if timestamp:
            row["timestamp"] = str(datetime.datetime.now())
        for param in query_list:
            value = fixed_to_float(node.sdo[param.name].raw)
            if symbolic and param.value_descriptions:
                if value in param.value_descriptions:
                    row_str = f"{param.value_descriptions[value]}"
                else:
                    row_str = f"{value:g} (Unknown value)"
            elif symbolic and param.bit_definitions:
                value = int(value)
                row_str = ""
                for bit, description in param.bit_definitions.items():
                    if bit & value:
                        row_str = row_str + description + ", "
                row_str = row_str.removesuffix(", ")

                if len(row_str) == 0:
                    row_str = "0"
            else:
                row_str = f"{value:g}"
            row[param.name] = row_str
        writer.writerow(row)
        out_file.flush()

        time.sleep(step)


@cli.command()
@click.argument("out_file", type=click.File("w"))
@pass_cli_settings
@db_action
@can_action
def save(cli_settings: CliSettings, out_file: click.File) -> None:
    """Save all parameters in json to OUT_FILE"""

    doc = {}
    count = 0
    node = cli_settings.node
    for item in cli_settings.database.names.values():
        if item.isparam:
            doc[item.name] = fixed_to_float(node.sdo[item.name].raw)
            count += 1

    json.dump(doc, out_file, indent=4)

    click.echo(f"Saved {count} parameters")


def set_enum_value(
        node: canopen.Node,
        param: OIVariable,
        value: str) -> None:
    """Set a enumeration parameter over SDO by looking up its symbolic value"""

    result = None
    for key, description in param.value_descriptions.items():
        if description.lower() == value.lower():
            result = key

    if result is not None:
        node.sdo[param.name].raw = fixed_from_float(result)
    else:
        click.echo(f"Unable to find value: '{value}' for parameter: "
                   f"{param.name}. Valid values are "
                   f"{param.value_descriptions}")


def set_bitfield_value(
        node: canopen.Node,
        param: OIVariable,
        value: str) -> None:
    """Set a bitfield parameter over SDO by looking up its symbolic values. The
      value should be a comma separated list"""

    result = 0
    for bit_name in value.split(','):
        bit_name = bit_name.strip()
        for key, description in param.bit_definitions.items():
            if description.lower() == bit_name.lower():
                result |= key

    node.sdo[param.name].raw = fixed_from_float(result)


def set_float_value(
        node: canopen.Node,
        param: OIVariable,
        value: float) -> None:
    """Set a parameter with a floating point value over SDO"""

    # pre-conditions that should always be
    assert param.isparam
    assert param.min is not None
    assert param.max is not None

    fixed_value = fixed_from_float(value)

    if fixed_value < param.min:
        click.echo(f"Value {value:g} is smaller than the minimum "
                   f"value {fixed_to_float(param.min):g} allowed "
                   f"for {param.name}")
    elif fixed_value > param.max:
        click.echo(f"Value {value:g} is larger than the maximum value "
                   f"{fixed_to_float(param.max):g} allowed for "
                   f"{param.name}")
    else:
        node.sdo[param.name].raw = fixed_value


def write_impl(
        cli_settings: CliSettings,
        param: str,
        value: Union[float, str]) -> None:
    """Implementation of the single parameter write command. Separated from
    the command so the logic can be shared with loading all parameters from
    json."""

    if param in cli_settings.database.names:
        param_item = cli_settings.database.names[param]

        # Check if we are a modifiable parameter
        if param_item.isparam:
            if isinstance(value, float):
                set_float_value(cli_settings.node, param_item, value)
            else:
                # Assume the value is a float to start with
                try:
                    set_float_value(
                        cli_settings.node,
                        param_item,
                        float(value))
                except ValueError:
                    if param_item.value_descriptions:
                        set_enum_value(cli_settings.node, param_item, value)
                    elif param_item.bit_definitions:
                        set_bitfield_value(
                            cli_settings.node, param_item, value)
                    else:
                        click.echo(f"Invalid value: '{value}' for parameter: "
                                   f"{param}")
        else:
            click.echo(f"{param} is a spot value parameter. "
                       "Spot values are read-only.")
    else:
        click.echo(f"Unknown parameter: {param}")


@cli.command(context_settings={"ignore_unknown_options": True})
@click.argument("param")
@click.argument("value")
@pass_cli_settings
@db_action
@can_action
def write(cli_settings: CliSettings, param: str, value: str) -> None:
    """Write the value to the parameter PARAM on the device"""

    write_impl(cli_settings, param, value)


@cli.command()
@click.argument("in_file", type=click.File("r"))
@pass_cli_settings
@db_action
@can_action
def load(cli_settings: CliSettings, in_file: click.File) -> None:
    """Load all parameters from json IN_FILE"""

    doc = json.load(in_file)
    count = 0
    for param_name in doc:
        value = doc[param_name]

        # JSON parameter files from openinverter.org/parameters/ put values in
        # quotation marks which isn't technically valid but we can work around
        if isinstance(value, str):
            value = float(value)

        write_impl(cli_settings, param_name, value)
        count += 1

    click.echo(f"Loaded {count} parameters")


@cli.command()
@pass_cli_settings
@can_action
def serialno(cli_settings: CliSettings) -> None:
    """Read the device serial number. This can be useful to recover failed
    devices using the 'upgrade' command."""

    assert cli_settings.node
    serialno_data = cli_settings.node.serial_no()
    part_str = []
    for part in range(0, 12, 4):
        part = serialno_data[part:part+4]
        part_str.append("".join(format(x, "02X") for x in part))

    click.echo(f"Serial Number: {part_str[0]}:{part_str[1]}:{part_str[2]}")


@cli.group()
@pass_cli_settings
def cmd(cli_settings: CliSettings) -> None:
    """Execute a command on a device"""
    # We have to have cli_settings to allow the command hierarchy to work but
    # it is unused here so just pretend to use it
    _ = cli_settings


@cmd.command("save")
@pass_cli_settings
@can_action
def cmd_save(cli_settings: CliSettings) -> None:
    """Save device parameters and CAN map to flash"""
    assert cli_settings.node
    cli_settings.node.save()
    click.echo("Command sent successfully")


@cmd.command("load")
@pass_cli_settings
@can_action
def cmd_load(cli_settings: CliSettings) -> None:
    """Load device parameters and CAN map from flash"""
    assert cli_settings.node
    cli_settings.node.load()
    click.echo("Command sent successfully")


@cmd.command("reset")
@pass_cli_settings
@can_action
def cmd_reset(cli_settings: CliSettings) -> None:
    """Reset the device"""
    assert cli_settings.node
    cli_settings.node.reset()
    click.echo("Command sent successfully")


@cmd.command("defaults")
@pass_cli_settings
@can_action
def cmd_defaults(cli_settings: CliSettings) -> None:
    """Reset the device parameters to their built-in defaults"""
    assert cli_settings.node
    cli_settings.node.load_defaults()
    click.echo("Command sent successfully")


@cmd.command("start")
@click.option("--mode",
              type=click.Choice(["Normal",
                                 "Manual",
                                 "Boost",
                                 "Buck",
                                 "ACHeat",
                                 "Sine"],
                                case_sensitive=False),
              default="Normal",
              show_default=True)
@pass_cli_settings
@can_action
def cmd_start(cli_settings: CliSettings, mode: str) -> None:
    """Start the device in the specified mode"""
    mode_list = {
        "Normal": oi.START_MODE_NORMAL,
        "Manual": oi.START_MODE_MANUAL,
        "Boost": oi.START_MODE_BOOST,
        "Buck": oi.START_MODE_BUCK,
        "Sine": oi.START_MODE_SINE,
        "ACHeat": oi.START_MODE_ACHEAT
    }

    assert cli_settings.node
    cli_settings.node.start(mode_list[mode])
    click.echo("Command sent successfully")


@cmd.command("stop")
@pass_cli_settings
@can_action
def cmd_stop(cli_settings: CliSettings) -> None:
    """Stop the device from operating"""
    assert cli_settings.node
    cli_settings.node.stop()
    click.echo("Command sent successfully")


@cli.group("can")
@pass_cli_settings
def can_map(cli_settings: CliSettings) -> None:
    """Manage parameter to CAN message mappings on a device"""
    # We have to have cli_settings to allow the command hierarchy to work but
    # it is unused here so just pretend to use it
    _ = cli_settings


def param_name_from_id(param_id: int, db: canopen.ObjectDictionary) -> str:
    """Return the name of a parameter based on the openinverter parameter ID.
    If it is not in the database the number is returned."""

    # This is not evenly remotely efficient
    param_name = None
    for item in db.names.values():
        if isinstance(item, OIVariable) and item.id == param_id:
            param_name = item.name
            break

    if param_name is None:
        return f"{param_id}"
    else:
        return f"{param_name}"


def print_can_map(
        direction_str: str,
        cur_map: List[CanMessage],
        db: canopen.ObjectDictionary) -> None:
    """Helper function to print the contents of a CAN message map """

    msg_index = 0
    param_index = 0
    for msg in cur_map:
        if msg.is_extended_frame:
            click.echo(f"{msg.can_id:#010x}:")
        else:
            click.echo(f"{msg.can_id:#x}:")

        for entry in msg.params:
            click.echo(
                f" {direction_str}.{msg_index}.{param_index}" +
                f" param='{param_name_from_id(entry.param_id, db)}'" +
                f" pos={entry.position} length={entry.length}" +
                f" gain={entry.gain} offset={entry.offset}"
            )
            param_index += 1
        param_index = 0
        msg_index += 1


@can_map.command("list")
@pass_cli_settings
@db_action
@can_action
def cmd_can_list(
    cli_settings: CliSettings,
) -> None:
    """List parameter to CAN message mappings"""

    assert cli_settings.node
    node = cli_settings.node

    tx_map = node.list_can_map(Direction.TX)
    rx_map = node.list_can_map(Direction.RX)

    if not tx_map and not rx_map:
        click.echo("(none)")
    else:
        print_can_map(
            "tx",
            tx_map,
            cli_settings.database)

        print_can_map(
            "rx",
            rx_map,
            cli_settings.database)


@can_map.command("add", context_settings={"ignore_unknown_options": True})
@click.argument("direction", required=True, type=click.Choice(["tx", "rx"]))
@click.argument("can_id", required=True)
@click.argument("param", required=True)
@click.argument("position", required=True, type=click.IntRange(0, 63))
@click.argument("length", required=True, type=click.IntRange(-32, 32))
@click.argument("gain",
                type=click.FloatRange(-8388.608, 8388.607),
                default=1.0)
@click.argument("offset", type=click.IntRange(-128, 127), default=0)
@click.option("--standard", "is_extended_frame", flag_value=False,
              default=True, show_default=True,
              help="Use standard format CAN 2.0a frames with 11-bit addresses "
              "(0-2047) or (0x000-0x7ff)")
@click.option("--extended", "is_extended_frame", flag_value=True,
              help="Use extended format CAN 2.0b frames with 29-bit addresses "
              "(0-536870911) or (0x00000000-0x1fffffff)")
@pass_cli_settings
@db_action
@can_action
def cmd_can_add(
    cli_settings: CliSettings,
    direction: str,
    can_id: str,
    param: str,
    position: int,
    length: int,
    gain: float,
    offset: int,
    is_extended_frame: bool
) -> None:
    """Add a CAN message mapping for a specific parameter

    \b
    CAN_ID   - A valid CAN ID in decimal or hexadecimal (prefix with '0x')
    PARAM    - The parameter name to map
    POSITION - The starting bit position (0-63)
    LENGTH   - The number of bits the value will take up (-32,-1) or (1,32)
    GAIN     - The gain to multiply the value (-8388.608, 8388.607)
    OFFSET   - The offset to add (-128, 127)

    For transmit mappings the parameter value is multiplied by the GAIN before
    adding the OFFSET. The value is then mapped into the specified bits.

    For receive mappings the value is identified in the CAN frame, the OFFSET
    subtracted before dividing by the GAIN.

    Positive LENGTH values indicate a little-endian mapping. The bits count
    up from the start POSITION.

    Negative LENGTH values indicate a big-endian mapping. The bits count down
    from the end POSITION.

    \b
    Example:
    $ oic -n22 can add tx 0x101 temp3 32 8
    """

    if length == 0:
        click.echo("Map length cannot be 0")
        return

    # Get the parameter id by name from the cached database
    if param in cli_settings.database.names:
        param_id = cast(OIVariable, cli_settings.database.names[param]).id
    else:
        click.echo(f"Parameter {param} not found in database")
        return

    # Parse both hex as 0x1337 -> 4919 and decimal as 1337 -> 1337
    can_id_int = literal_eval(can_id)
    if is_extended_frame:
        if can_id_int < 0 or can_id_int >= 0x20000000:
            click.echo(f"can_id: {can_id} out of range for an extended CAN " +
                       "frame.\nExpected 0x00000000-0x1fffffff or 0-536870911")
            return
    else:
        if can_id_int < 0 or can_id_int >= 0x800:
            click.echo(f"can_id: {can_id} out of range for a standard CAN " +
                       "frame.\nExpected 0x000-0x7ff or 0-2047")
            return

    click.echo(f"Adding CAN {direction} mapping with ", nl=False)
    if is_extended_frame:
        click.echo(f"can_id={can_id_int:#010x} ", nl=False)
    else:
        click.echo(f"can_id={can_id_int:#x} ", nl=False)
    click.echo(f"param='{param}' position={position} " +
               f"length={length} gain={gain} offset={offset}")

    assert cli_settings.node
    node = cli_settings.node
    node.add_can_map_entry(
        can_id_int,
        Direction[direction.upper()],
        param_id,
        position,
        length,
        gain,
        offset,
        is_extended_frame)

    click.echo("CAN mapping added successfully.")


@can_map.command("remove")
@click.argument("listing_id", required=True, type=str)
@pass_cli_settings
@can_action
def cmd_can_remove(
    cli_settings: CliSettings,
    listing_id: str,
) -> None:
    """
    Remove a parameter to CAN message mapping

    Mapping is referred to by an id in the format [tx/rx].<n>.<n>.

    \b
    Example:
    $ oic can list
    0x202:
     rx.0.0 param='soc' pos=16 len=8 gain=1.0 offset=0
    0x203:
     rx.1.0 param='maxpower' pos=32 len=16 gain=1.0 offset=0
     rx.1.2 param='maxcur' pos=16 len=16 gain=1.0 offset=0
    $ oic can remove rx.1.2
    CAN mapping removed successfully.
    $ oic can remove rx.0.0
    CAN mapping removed successfully.
    $ oic can list
    0x203:
     rx.1.0 param='maxpower' pos=32 len=16 gain=1.0 offset=0
    """

    listing_parts = re.match(
        r"^(tx|rx)\.(\d{1,2})\.(\d{1,2})$",
        listing_id,
        flags=re.IGNORECASE)

    if not listing_parts:
        click.echo('Invalid listing id: Correct format is "[tx/rx].<n>.<n>"')
        return

    direction = Direction[listing_parts[1].upper()]
    msg_index = int(listing_parts[2])
    param_index = int(listing_parts[3])

    assert cli_settings.node
    node = cli_settings.node
    if node.remove_can_map_entry(
            direction,
            msg_index,
            param_index):
        click.echo("CAN mapping removed successfully.")
    else:
        click.echo("Unable to find CAN map entry.")


@can_map.command("clear")
@click.argument("direction",
                required=True, type=click.Choice(["tx", "rx", "all"]),
                default="all")
@pass_cli_settings
@can_action
def cmd_can_clear(
    cli_settings: CliSettings,
    direction: str,
) -> None:
    """
    Clear all parameter to CAN message mappings for the specified direction
    """

    assert cli_settings.node
    node = cli_settings.node
    if direction == "all":
        node.clear_map(Direction.TX)
        node.clear_map(Direction.RX)
    else:
        node.clear_map(Direction[direction.upper()])

    click.echo(f"CAN {direction} mapping removed successfully.")


@can_map.command("export")
@click.argument("out_file",
                type=click.Path(
                    file_okay=True,
                    dir_okay=False,
                    writable=True,
                    path_type=Path))
@click.option("--format",
              type=click.Choice(["json",
                                 "dbc"],
                                case_sensitive=False),
              default="json",
              show_default=True)
@pass_cli_settings
@db_action
@can_action
def cmd_can_export(
    cli_settings: CliSettings,
    out_file: Path,
    format: str  # pylint: disable=redefined-builtin
) -> None:
    """Export all parameter to CAN message mappings to OUT_FILE"""

    assert cli_settings.node
    node = cli_settings.node

    tx_map = node.list_can_map(Direction.TX)
    rx_map = node.list_can_map(Direction.RX)

    if format == "json":
        with open(out_file, "wt", encoding="utf-8") as json_file:
            export_json_map(tx_map, rx_map, cli_settings.database, json_file)

    elif format == "dbc":
        export_dbc_map(f"node{cli_settings.node_number}",
                       tx_map, rx_map, cli_settings.database, out_file)

    click.echo("Parameter CAN message map exported")


@can_map.command("import")
@click.argument("in_file", type=click.File("r"))
@click.option("--clear/--no-clear",
              show_default=True,
              default=True,
              help="Clear any existing CAN message map")
@pass_cli_settings
@db_action
@can_action
def cmd_can_import(cli_settings: CliSettings,
                   in_file: click.File,
                   clear: bool) -> None:
    """Import a CAN message map from a json IN_FILE"""

    assert cli_settings.node
    node = cli_settings.node

    if clear:
        node.clear_map(Direction.TX)
        node.clear_map(Direction.RX)
        click.echo("Existing CAN message map cleared")

    (tx_map, rx_map) = import_json_map(in_file, cli_settings.database)

    node.add_can_map(Direction.TX, tx_map)
    click.echo("Transmit CAN message map configured")

    node.add_can_map(Direction.RX, rx_map)
    click.echo("Receive CAN message map configured")


@cli.command()
@pass_cli_settings
@can_action
def scan(cli_settings: CliSettings) -> None:
    """Scan the CAN bus for available nodes"""

    assert cli_settings.network is not None

    # Maximum number of devices we are going to scan for
    limit = 127

    # Canned SDO request we use to scan the bus with to find something
    # we can talk to
    sdo_req = b"\x40\x00\x10\x00\x00\x00\x00\x00"

    click.echo("Scanning for devices. Please wait...\n")

    # Implement our own scanner rather than use
    # canopen.network.scanner.search() as this lets us rate limit the scan to
    # avoid exhausting local CAN network queues
    for node_id in range(1, limit + 1):
        cli_settings.network.send_message(0x600 + node_id, sdo_req)
        time.sleep(0.01)

    # Wait for any responses to show up
    time.sleep(5)

    # filter out weird canopen internal node IDs that show up here and
    # nowhere else
    node_list = [id for id in cli_settings.network.scanner.nodes if id < limit]

    if node_list:
        for node_id in node_list:
            click.echo(f"Found possible openinverter node: {node_id}")
    else:
        click.echo("No nodes found")


@cli.command()
@pass_cli_settings
@can_action
@click.argument("firmware_file",
                type=click.Path(
                    file_okay=True,
                    dir_okay=False,
                    writable=False,
                    path_type=Path))
@click.option("-s", "--serial",
              default=None,
              type=click.STRING,
              help="The serial number of a specific device to recover. "
              "Only the first 8 digits of the serial number are required.")
@click.option("--recover",
              is_flag=True,
              help="Recover a device as it starts up.")
@click.option("--wait",
              default=5.0,
              show_default=True,
              type=click.FLOAT,
              help="Time to wait for a device to reset and start the upgrade "
              " process in seconds")
def upgrade(
        cli_settings: CliSettings,
        firmware_file: Path,
        serial: str,
        recover: bool,
        wait: float) -> None:
    """
    Upgrade the device firmware.

    For devices that are operating normally the process is fully automatic.

    It is possible to recover a faulty device. The upgrade process will wait
    until the device boots before starting the upgrade process. It is up to
    the user to boot the device, typically by powering it on.

    On a CAN network with more than one device it is recommended to specify
    the device serial number. The upgrade process will ensure that only this
    device is upgraded. If no serial number is provided the first device to
    boot will be upgraded.
    """

    failure_messages = {
        Failure.PROTOCOL_ERROR:
        "Unexpected CAN frame received from device",

        Failure.UPGRADE_IN_PROGRESS:
        "An upgrade is already in progress on the CAN bus",

        Failure.PAGE_CRC_ERROR:
        "Firmware upload data corruption detected"
    }

    def _print_progress(update: StateUpdate) -> None:
        if update.state == State.START:
            click.echo("Waiting for device to connect...", nl=False)

        elif update.state == State.HEADER:
            serialno_str = "".join(format(x, "02x") for x in update.serialno)
            click.echo(f"\rDevice upgrade started for {serialno_str}")

        elif update.state in (State.UPLOAD, State.CHECK_CRC):
            progress = update.progress
            click.echo(f"\rUpgrading: {progress:.1f}% complete", nl=False)

        elif update.state == State.WAIT_FOR_DONE:
            click.echo(
                "\rWaiting for device to complete upgrade", nl=False)

        elif update.state == State.FAILURE:
            if update.failure in failure_messages:
                click.echo(
                    f"Upgrade failed: {failure_messages[update.failure]}")
            else:
                click.echo(
                    f"Upgrade failed: Unknown failure - {update.failure}")

        elif update.state == State.COMPLETE:
            click.echo("\rUpgrade completed successfully!".ljust(40))

    if recover:
        if serial and len(serial) != 8:
            click.echo("Device serial numbers should be 8 hexadecimal digits")
            return

        if serial:
            recovery_serialno = bytes.fromhex(serial)
        else:
            recovery_serialno = None

        upgrader = CanUpgrader(cli_settings.network, recovery_serialno,
                               firmware_file, _print_progress)

    else:
        if serial:
            click.echo("Serial numbers do not need to be provided for normal "
                       "upgrades")
            return

        assert cli_settings.node
        node_serialno = cli_settings.node.serial_no()

        upgrader = CanUpgrader(cli_settings.network, node_serialno[:4],
                               firmware_file, _print_progress)

        try:
            if not cli_settings.debug:
                # suppress logging errors from the canopen library if the
                # device doesn't respond when asked to reset
                logging.disable(logging.ERROR)

            cli_settings.node.reset()
        except canopen.SdoCommunicationError:
            pass

    if not upgrader.run(wait):
        click.echo("\r\nUpgrade timed out")


@cli.group()
@pass_cli_settings
def cache(cli_settings: CliSettings) -> None:
    """Parameter database cache management commands"""
    # We have to have cli_settings to allow the command hierarchy to work but
    # it is unused here so just pretend to use it
    _ = cli_settings


@cache.command("clean")
@pass_cli_settings
def cmd_clean(cli_settings: CliSettings) -> None:
    """Remove all entries from the parameter database cache"""

    # cli_settings is unused
    _ = cli_settings

    count = 0
    for file in glob.glob(
            os.path.join(
                appdirs.user_cache_dir(oi.APPNAME, oi.APPAUTHOR), "*.json")):
        click.echo(f"Removing {file}")
        os.remove(file)
        count += 1

    if count == 0:
        click.echo("No cache entries found")



"""openinverter remote database unit tests"""
import csv
import sys
import tempfile
import unittest
from pathlib import Path

import pytest

from openinverter_can_tool.paramdb import import_database
from openinverter_can_tool.remote_db import RemoteDatabaseNode

from .network_test_case import NetworkTestCase

TX = 1
RX = 2

CAPTURE_DATA_DIR = Path(__file__).parent / "test_data" / "captures"

# Reduce test verbosity
# pylint: disable=missing-function-docstring


class TestRemoteDatabaseNode(NetworkTestCase):
    """
    Test the openinverter specific database access SDO indices
    """

    def __init__(self, methodName: str = "runTest") -> None:
        super().__init__(methodName)
        self._node_type = RemoteDatabaseNode

    def test_paramdb_checksum(self):
        self.data = [
            (TX, b'\x40\x00\x50\x03\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x50\x03\x12\x70\x01\x00')
        ]
        checksum = self.node.param_db_checksum()
        assert checksum == 94226

    def load_capture(self, capture: str, tx_id: int, rx_id: int) -> None:
        """Load a CAN capture from SavvyCAN in CSV format into the data"""
        with open(CAPTURE_DATA_DIR / capture,
                  newline="", encoding="utf-8") as capture_file:
            reader = csv.DictReader(capture_file)

            for row in reader:
                frame_id = int(row["ID"], 16)

                if frame_id == tx_id:
                    direction = TX
                elif frame_id == rx_id:
                    direction = RX
                else:
                    continue

                assert int(row["LEN"]) == 8

                frame = row["D1"] + row["D2"] + row["D3"] + row["D4"] + \
                    row["D5"] + row["D6"] + row["D7"] + row["D8"]
                self.data.append((direction, bytes.fromhex(frame)))

    @pytest.mark.skipif(sys.version_info < (3, 12),
                        reason="NamedTemporaryFile() missing delete_on_close "
                        "before python 3.12")
    def test_query_zombieverter_paramdb(self):
        self.load_capture(
            "zombieverter-node3-query-paramdb.csv",
            0x603,
            0x583)
        self.node.node_id = 3

        checksum = self.node.param_db_checksum()

        with tempfile.NamedTemporaryFile(delete_on_close=False) as db_file:
            db_file.write(self.node.param_db())
            db_file.close()

            db = import_database(Path(db_file.name))

        assert checksum == 181129
        assert len(db.names) == 194


if __name__ == "__main__":
    unittest.main()



"""
openinverter remote database access
"""

import struct

import canopen
import canopen.objectdictionary
from canopen.sdo import SdoClient

from . import constants as oi


class RemoteDatabaseNode:
    """A simplified CANopen SDO wrapper around the two indexes that implement
    the remote database definition."""

    def __init__(self, network: canopen.Network, node_id: int) -> None:
        """Create temporary SDO client and attach to the network """

        self.network = network
        self.node_id = node_id
        self.sdo_client = SdoClient(0x600 + node_id,
                                    0x580 + node_id,
                                    canopen.ObjectDictionary())
        self.sdo_client.network = network  # type: ignore
        network.subscribe(0x580 + node_id, self.sdo_client.on_response)

    def __del__(self) -> None:
        can_id = 0x580 + self.node_id

        if can_id in self.network.subscribers:
            self.network.unsubscribe(can_id)

    def param_db_checksum(self) -> int:
        """
        Read the parameter database checksum. If this is equal to a
        previously read value the bytes read from ParamDb() can be
        considered equal. A different value implies that any data read
        from the ParamDb() method should be discarded.
        """
        value, = struct.unpack("<L",
                               self.sdo_client.upload(
                                   oi.SERIALNO_INDEX,
                                   oi.PARAM_DB_CHECKSUM_SUBINDEX))

        return value

    def param_db(self) -> bytes:
        """Read the remote parameter database"""
        with self.sdo_client.open(oi.STRINGS_INDEX,
                                  oi.PARAM_DB_SUBINDEX,
                                  "rb") as param_db:
            return param_db.read()  # type: ignore



"""
Routines to allow CAN message maps to be persisted
"""

import json
from collections import OrderedDict
from pathlib import Path
from typing import IO, Dict, List, Tuple, Optional

import canopen.objectdictionary
import cantools
import cantools.database

from .fpfloat import fixed_to_float
from .oi_node import CanMessage, MapEntry
from .paramdb import OIVariable


def export_json_map(tx_map: List[CanMessage],
                    rx_map: List[CanMessage],
                    db: canopen.ObjectDictionary,
                    out_file: IO) -> None:
    """
    Export the provided CAN message maps encoded as JSON into the specified
    file

    :param tx_map:  The transmit CAN message map
    :param rx_map:  The receive CAN message map
    :param db:      The object database to convert parameter IDs to names with
    :param out_file: The writeable file object to output the encoded JSON to
    """

    def _convert_map_to_dict(msg_map: List[CanMessage]) -> List[Dict]:
        out_list = []
        for msg in msg_map:
            out_params = []
            out_msg = {
                "can_id": msg.can_id,
                "params": out_params,
                "is_extended_frame": msg.is_extended_frame
            }
            for entry in msg.params:
                # Search inefficiently for the parameter name
                param_name = None
                for item in db.names.values():
                    if isinstance(item, OIVariable) and \
                            item.id == entry.param_id:
                        param_name = item.name
                        break

                if param_name is None:
                    raise KeyError(entry.param_id)

                out_params.append({
                    "param": param_name,
                    "position": entry.position,
                    "length": entry.length,
                    "gain": entry.gain,
                    "offset": entry.offset
                })
            out_list.append(out_msg)

        return out_list

    doc = {
        "version": 2,
        "tx": _convert_map_to_dict(tx_map),
        "rx": _convert_map_to_dict(rx_map)
    }
    json.dump(doc, out_file, indent=4)


def import_json_map(in_file: IO,
                    db: canopen.ObjectDictionary,
                    ) -> Tuple[List[CanMessage],
                               List[CanMessage]]:
    """
    Import a CAN message map from the supplied JSON file.

    :param in_file: The JSON file with the CAN message map to import
    :param db:      The object database to resolve parameter names with

    :returns: Tuple with the transmit and receive CAN message maps
    """

    def _parse_map_entries(params_doc) -> List[MapEntry]:
        params = []
        for param in params_doc:
            param_id = db.names[param["param"]].id
            params.append(
                MapEntry(
                    param_id,
                    param["position"],
                    param["length"],
                    param["gain"],
                    param["offset"],
                ))

        return params

    def _parse_can_messages(msg_doc) -> List[CanMessage]:
        msg_list = []
        for msg in msg_doc:
            if "is_extended_frame" in msg:
                is_extended_frame = msg["is_extended_frame"]
            else:
                is_extended_frame = False
            msg_list.append(
                CanMessage(msg["can_id"],
                           _parse_map_entries(msg["params"]),
                           is_extended_frame))
        return msg_list

    doc = json.load(in_file)

    if "version" not in doc or "tx" not in doc or "rx" not in doc:
        raise RuntimeError("Invalid file format")

    version = doc["version"]
    if version not in (1, 2):
        # Version 1: Original version introduced in 0.0.9 release
        # Version 2: Adds is_extended_frame field to the CAN message
        raise RuntimeError(f"Unsupported version: {version}")

    tx_map = _parse_can_messages(doc["tx"])
    rx_map = _parse_can_messages(doc["rx"])

    return (tx_map, rx_map)


# cantools don't export the Database properly but it is safe to use this type
# pyright: reportPrivateImportUsage=false

def transform_map_to_canopen_db(
        node_prefix: Optional[str],
        tx_map: List[CanMessage],
        rx_map: List[CanMessage],
        db: canopen.ObjectDictionary) -> cantools.database.can.Database:
    """
    Transform the provided CAN message maps to a canopen database

    :param node_prefix: An optional string prefix that is added to each db node
    :param tx_map:      The transmit CAN message map
    :param rx_map:      The receive CAN message map
    :param db:          The object database to convert parameter IDs to names
                        with

    :returns: The canopen database representing the two maps
    """

    def _find_param(param_id: int) -> OIVariable:
        """Search inefficiently for the parameter given the openinverter
        internal id"""

        for item in db.names.values():
            if isinstance(item, OIVariable) and item.id == param_id:
                return item

        # Didn't find the parameter
        raise KeyError(param_id)

    def _convert_param_to_signal(
        param_name: str,
        param: OIVariable,
        entry: MapEntry,
        is_signed: bool
    ) -> cantools.database.can.signal.Signal:
        if entry.length > 0:
            byte_order = "little_endian"
            start_bit = entry.position
            bit_length = entry.length
        else:
            byte_order = "big_endian"
            start_bit = entry.position + entry.length + 8
            bit_length = -entry.length

        signal = cantools.database.can.signal.Signal(
            name=param_name,
            start=start_bit,
            length=bit_length,
            byte_order=byte_order
        )

        if param.value_descriptions:
            values = OrderedDict()
            for value, description in param.value_descriptions.items():
                values[int(value*entry.gain+entry.offset)] = description
            signal.choices = values
            signal.is_signed = False

        elif param.bit_definitions:
            bits = OrderedDict()
            for value, description in param.bit_definitions.items():
                bits[int(value*entry.gain+entry.offset)] = description
            signal.choices = bits
            signal.is_signed = False

        else:
            # dbc files scale and offset are the inverse of openinverter
            # gain and offset
            signal.scale = 1.0 / entry.gain
            signal.offset = -entry.offset
            signal.is_signed = is_signed

            if param.isparam:
                signal.unit = param.unit

        if param.isparam:
            if param.min:
                signal.minimum = fixed_to_float(param.min)
            if param.max:
                signal.maximum = fixed_to_float(param.max)

        return signal

    def _convert_map_to_messages(
            msg_map: List[CanMessage],
            node_name: str,
            msg_prefix: str,
            is_signed: bool
    ) -> List[cantools.database.can.message.Message]:
        out_list = []
        msg_no = 1
        for msg in msg_map:
            signals = []
            signal_names = {}
            for entry in msg.params:
                param = _find_param(entry.param_id)

                # Ensure we don't have duplicate signal names
                param_name = param.name
                if param_name in signal_names:
                    signal_names[param_name] += 1
                    param_name = f"{param_name}_{signal_names[param_name]}"
                else:
                    signal_names[param_name] = 0

                signals.append(
                    _convert_param_to_signal(
                        param_name, param, entry, is_signed)
                )

            out_msg = cantools.database.can.message.Message(
                name=f"{msg_prefix}_msg{msg_no}",
                frame_id=msg.can_id,
                is_extended_frame=msg.is_extended_frame,
                length=8,
                signals=signals,
                senders=[node_name]
            )
            out_list.append(out_msg)
            msg_no += 1

        return out_list

    tx_node = cantools.database.can.node.Node(
        f"{node_prefix}_tx" if node_prefix else "tx"
    )
    rx_node = cantools.database.can.node.Node(
        f"{node_prefix}_rx" if node_prefix else "rx"
    )

    nodes = []
    if tx_map:
        nodes.append(tx_node)
    if rx_map:
        nodes.append(rx_node)

    messages = _convert_map_to_messages(tx_map, tx_node.name, "tx", True)
    messages += _convert_map_to_messages(rx_map, rx_node.name, "rx", False)

    return cantools.database.can.Database(messages, nodes)


def export_dbc_map(
        node_prefix: Optional[str],
        tx_map: List[CanMessage],
        rx_map: List[CanMessage],
        db: canopen.ObjectDictionary,
        out_file: Path) -> None:
    """
    Export the provided CAN message maps to a DBC

    :param node_prefix: An optional string prefix that is added to each db node
    :param tx_map:      The transmit CAN message map
    :param rx_map:      The receive CAN message map
    :param db:          The object database to convert parameter IDs to names
                        with
    :param out_file:    The file path to output the DBC to
    """
    cantools_db = transform_map_to_canopen_db(node_prefix, tx_map, rx_map, db)

    cantools.database.dump_file(cantools_db, out_file)



"""
Functions to manipulate openinverter fixed point values
"""

FACTOR = int(2**5)


def fixed_from_float(value: float) -> int:
    """convert a floating point value to a 32-bit/5-bit fixed point value"""
    return int(value * FACTOR)


def fixed_to_float(value: int) -> float:
    """convert a 32-bit/5-bit fixed point value to fixed point value"""
    return float(value / float(FACTOR))



"""openinverter CAN upgrade module"""

from __future__ import annotations

import array
import queue
import struct
from abc import ABC, abstractmethod
from collections import namedtuple
from enum import IntEnum, auto
from pathlib import Path
from typing import List, Optional, Callable

import canopen

DEVICE_CAN_ID = 0x7DE
UPGRADER_CAN_ID = 0x7DD

PAGE_SIZE = 1024

MAX_PAGES = 255  # Most we can fit in the single-byte response to 'S'


class DevicePacket(IntEnum):
    """Different upgrade packet signatures sent by devices"""
    HELLO = 0x33  # '3'
    START = 0x53  # 'S'
    PAGE = 0x50   # 'P'
    CRC = 0x43    # 'C'
    DONE = 0x44   # 'D'
    ERROR = 0x45  # 'E'


def stm_crc(data: bytes):
    """
    Compute the CRC-32 expected by the stm32-CANBootloader firmware.

    This requires:
     - mapping the data to little-endian unsigned integers
     - Using the CRC-32/MPEG-2 with:
       - width=32
       - poly=0x04c11db7
       - init=0xffffffff
       - refin=false
       - refout=false
    """

    crc = 0xffffffff
    data_array = array.array("I", data)
    for word in data_array:
        crc = crc ^ word

        for _ in range(32):
            if crc & 0x80000000:
                # Polynomial used in STM32
                crc = ((crc << 1) ^ 0x04C11DB7) & 0xffffffff
            else:
                crc = (crc << 1) & 0xffffffff
    return crc


class Page:
    """A firmware page"""

    def __init__(self, data: bytes) -> None:
        assert len(data) <= PAGE_SIZE

        self.data = data
        padding = PAGE_SIZE - len(data) % PAGE_SIZE
        if padding != PAGE_SIZE:
            self.data += bytes(padding)

        self.crc = stm_crc(self.data)


class State(IntEnum):
    """List of possible states the upgrade process can be in"""
    START = auto()
    HEADER = auto()
    UPLOAD = auto()
    CHECK_CRC = auto()
    WAIT_FOR_DONE = auto()
    FAILURE = auto()
    COMPLETE = auto()


StateUpdate = namedtuple("StateUpdate", "state serialno failure progress")


class CanUpgrader:

    """
    Class to manage upgrading the firmware of open inverter devices over a
    CAN bus connection.

    :param network: The canopen.Network used to communicate with the device.
    :param serialno: The serial number of the device to look for booting. If
                     not specified the next device to boot will be upgraded.
    :param firmware: Path to the firmware file to upload.
    :param callback: Optional callback function to receive state changes.
    """

    def __init__(
            self,
            network: canopen.Network,
            serialno: Optional[bytes],
            firmware: Path,
            callback: Optional[Callable[[StateUpdate]]] = None):
        self._callback = callback
        self._pages: List[Page] = []

        with open(firmware, "rb") as firmware_file:
            while True:
                data = firmware_file.read(PAGE_SIZE)

                if data:
                    self.pages.append(Page(data))
                    if len(self.pages) > MAX_PAGES:
                        raise ValueError("Firmware image too large")
                else:
                    break

        self._status_queue = queue.Queue()

        self._state: State = State.START
        self._serialno: Optional[bytes] = None
        self._failure: Optional[Failure] = None
        self._progress: float = 0.0

        self._state_machine = StateMachine(network,
                                           serialno,
                                           self._pages,
                                           self._status_queue)

    def run(self, timeout: float) -> bool:
        """
        Run the upgrade process to completion or until the timeout is
        exceeded.

        :param timeout: The time to wait for the process to timeout in seconds

        :returns: True if the process ran to a final state (CompleteState or
                  FailureState) or False if the process timed out
        """
        try:
            while True:
                update = self._status_queue.get(True, timeout)

                self._state = update.state
                self._serialno = update.serialno
                self._failure = update.failure
                self._progress = update.progress

                if self._callback:
                    self._callback(update)

                if self._state in (State.COMPLETE, State.FAILURE):
                    return True
        except queue.Empty:
            return False

    @property
    def state(self) -> State:
        """Retrieve the current state of the upgrade process."""
        return self._state

    @property
    def failure(self) -> Optional[Failure]:
        """Return the failure details"""
        return self._failure

    @property
    def progress(self) -> float:
        """Return the percentage progress through the upgrade process"""
        return self._progress

    @property
    def serialno(self) -> Optional[bytes]:
        """
        Retrieve the serial number of the current device being upgraded.
        This is only valid once the HEADER has been successfully achieved.
        """
        return self._serialno

    @property
    def pages(self) -> List[Page]:
        """The list of loaded firmware pages to be uploaded to the device"""
        return self._pages


class StateMachine:
    """
    Internal state-machine managing the upgrade process. After
    initialisation this assumed to be called from a separate notification
    from the CAN interface.
    :param network: The canopen.Network to communicate with the device
    :param serialno: The optional serial number to match
    :param pages: The list of firmware pages to upload to the device
    :param status_queue: The queue to send status updates back to CanUpgrader
    """

    _state: InternalState

    def __init__(
            self,
            network: canopen.Network,
            serialno: Optional[bytes],
            pages: List[Page],
            status_queue: queue.Queue) -> None:
        self._network = network
        self._target_serialno = serialno
        self._serialno: Optional[bytes] = None
        self._pages = pages

        self._current_page = 0
        self._confirmed_page = -1
        self._status_queue = status_queue

        self.transition_to(StartState())

        # Once we subscribe data will start arriving on another thread
        network.subscribe(DEVICE_CAN_ID, self._process_can_frame)

    def transition_to(self, state: InternalState) -> None:
        """
        Allow states to transition to a new state
        """
        self._state = state
        self._state.sm = self
        self._status_queue.put(state.details())

    @property
    def serialno(self) -> Optional[bytes]:
        """
        Retrieve the serial number of the current device being upgraded.
        This is only valid once the HeaderState has been successfully
        achieved.
        """
        return self._serialno

    @serialno.setter
    def serialno(self, serialno: bytes) -> None:
        self._serialno = serialno

    @property
    def target_serialno(self) -> Optional[bytes]:
        """
        Retrieve the target serial number the upgrade process is directed
        towards.
        """
        return self._target_serialno

    @property
    def pages(self) -> List[Page]:
        """The list of loaded firmware pages to be uploaded to the device"""
        return self._pages

    @property
    def current_page(self) -> Page:
        """Return the current firmware page. May raise IndexError."""
        return self._pages[self._current_page]

    def advance_page(self) -> None:
        """Move to the next firmware page."""
        self._current_page += 1

    def confirm_page(self) -> None:
        """Confirm correct reception of a previous firmware page."""
        self._confirmed_page += 1

    def progress(self) -> float:
        """Compute the internal upgrade progress percentage."""
        if self._confirmed_page < 0:
            return 0.0

        if len(self._pages) > 0:
            return (self._confirmed_page * 100.0) / len(self._pages)
        else:
            return 100.0

    def reply(self, data: bytes) -> None:
        """Send a reply to the device."""
        self._network.send_message(UPGRADER_CAN_ID, data)

    def _process_can_frame(self,
                           can_id: int,
                           data: bytearray,
                           timestamp: float) -> None:
        """Process an incoming CAN frame"""
        assert can_id == DEVICE_CAN_ID
        _ = timestamp
        self._state.process(bytes(data))


class InternalState(ABC):
    """
    The base State class declares methods that all concrete states should
    implement and also provides a backreference to the StateMachine object,
    associated with the State. This backreference can be used by States to
    transition the StateMachine to another State.
    """
    sm: StateMachine

    @abstractmethod
    def process(self, data: bytes) -> None:
        """Process a new CAN message"""

    @abstractmethod
    def details(self) -> StateUpdate:
        """Return the details of the current state suitable for external
        consumption"""


class StartState(InternalState):
    """Waiting for a valid magic frame indicating a device is booting"""

    def process(self, data: bytes) -> None:
        # Recognise valid HELLO frames
        if len(data) == 8 and data[0] == DevicePacket.HELLO:
            # Reverse the bytes on the wire before
            device_serialno = data[7:3:-1]
            if (self.sm.target_serialno is None or
                    device_serialno == self.sm.target_serialno):
                self.sm.serialno = device_serialno
                self.sm.transition_to(HeaderState())
                self.sm.reply(data[4:8])

        # Explicitly reject any frames that could come from a device in the
        # middle of an upgrade process
        elif len(data) == 1 and data[0] in (DevicePacket.START,
                                            DevicePacket.PAGE,
                                            DevicePacket.CRC,
                                            DevicePacket.DONE):
            self.sm.transition_to(
                FailureState(Failure.UPGRADE_IN_PROGRESS))
        else:
            # Any other data on this CAN ID indicates a device which might
            # corrupt the upgrade process
            self.sm.transition_to(FailureState(Failure.PROTOCOL_ERROR))

    def details(self) -> StateUpdate:
        return StateUpdate(State.START, self.sm.serialno, None, 0)


class HeaderState(InternalState):
    """Wait for the device to request details of the firmware image"""

    def process(self, data: bytes) -> None:
        if len(data) == 1 and data[0] == DevicePacket.START:
            try:
                self.sm.transition_to(
                    UploadState(self.sm.current_page))
            except IndexError:
                self.sm.transition_to(WaitForDoneState())

            self.sm.reply(
                struct.pack("B", len(self.sm.pages)))
        elif len(data) == 8 and data[0] == DevicePacket.HELLO:
            pass
        else:
            self.sm.transition_to(FailureState(Failure.PROTOCOL_ERROR))

    def details(self) -> StateUpdate:
        return StateUpdate(State.HEADER, self.sm.serialno, None, 0)


class UploadState(InternalState):
    """Wait for the device to request a new page of firmware data to upload"""

    def __init__(self, page: Page) -> None:
        self.page = page
        self.pos = 0

    def process(self, data: bytes) -> None:
        if len(data) == 1 and data[0] == DevicePacket.PAGE:
            if self.pos == 0:
                self.sm.confirm_page()

            self.sm.reply(self.page.data[self.pos:self.pos+8])
            self.pos += 8

            if self.pos == PAGE_SIZE:
                self.sm.transition_to(CheckCrcState(self.page.crc))

        elif len(data) == 1 and data[0] == DevicePacket.ERROR:
            self.sm.transition_to(FailureState(Failure.PAGE_CRC_ERROR))

        elif len(data) == 8 and data[0] == DevicePacket.HELLO:
            pass
        else:
            self.sm.transition_to(FailureState(Failure.PROTOCOL_ERROR))

    def details(self) -> StateUpdate:
        return StateUpdate(
            State.UPLOAD,
            self.sm.serialno,
            None,
            self.sm.progress())


class CheckCrcState(InternalState):
    """Waiting for the device to validate the CRC of a completed page"""

    def __init__(self, crc: int) -> None:
        self.crc = crc

    def process(self, data: bytes) -> None:
        if len(data) == 1 and data[0] == DevicePacket.CRC:
            self.sm.reply(struct.pack("<I", self.crc))

            self.sm.advance_page()
            try:
                self.sm.transition_to(
                    UploadState(self.sm.current_page))
            except IndexError:
                self.sm.transition_to(WaitForDoneState())

        elif len(data) == 8 and data[0] == DevicePacket.HELLO:
            pass
        else:
            self.sm.transition_to(FailureState(Failure.PROTOCOL_ERROR))

    def details(self) -> StateUpdate:
        return StateUpdate(
            State.CHECK_CRC,
            self.sm.serialno,
            None,
            self.sm.progress())


class WaitForDoneState(InternalState):
    """Waiting until the device confirms completion of the upgrade"""

    def process(self, data: bytes) -> None:
        if len(data) == 1 and data[0] == DevicePacket.DONE:
            self.sm.confirm_page()
            self.sm.transition_to(CompleteState())

        elif len(data) == 1 and data[0] == DevicePacket.ERROR:
            self.sm.transition_to(FailureState(Failure.PAGE_CRC_ERROR))

        elif len(data) == 8 and data[0] == DevicePacket.HELLO:
            pass
        else:
            self.sm.transition_to(FailureState(Failure.PROTOCOL_ERROR))

    def details(self) -> StateUpdate:
        return StateUpdate(
            State.WAIT_FOR_DONE,
            self.sm.serialno,
            None,
            self.sm.progress())


class Failure(IntEnum):
    """Failure code for the upgrade process"""
    PROTOCOL_ERROR = 1
    UPGRADE_IN_PROGRESS = 2
    PAGE_CRC_ERROR = 3


class FailureState(InternalState):
    """The firmware upgrade process has failed. This is a final state for the
    upgrade state machine."""

    def __init__(self, failure: Failure) -> None:
        self.failure = failure

    def process(self, data: bytes) -> None:
        pass

    def details(self) -> StateUpdate:
        return StateUpdate(
            State.FAILURE,
            self.sm.serialno,
            self.failure,
            self.sm.progress())


class CompleteState(InternalState):
    """The firmware upgrade process is complete. This is a final state for the
    upgrade state machine."""

    def process(self, data: bytes) -> None:
        pass

    def details(self) -> StateUpdate:
        return StateUpdate(
            State.COMPLETE,
            self.sm.serialno,
            None,
            self.sm.progress())



"""
openinverter specific CANopen API
"""

import struct
from enum import IntEnum
from typing import List, Optional

import canopen
import canopen.objectdictionary
from canopen.node.base import BaseNode
from canopen.sdo import SdoClient

from . import constants as oi

# Common data type
UNSIGNED32 = struct.Struct("<L")


class Direction(IntEnum):
    """Direction for mapped parameters"""
    TX = 1
    RX = 2


MapListDirectionIndex = {
    Direction.TX: oi.CAN_MAP_LIST_TX_INDEX,
    Direction.RX: oi.CAN_MAP_LIST_RX_INDEX
}


def _validate_map_entry_parameters(
        position: int,
        length: int,
        gain: float,
        offset: int) -> None:
    """Common validation of map entry parameters"""
    if position not in range(0, 64):
        raise ValueError
    if abs(length) not in range(1, 33):
        raise ValueError
    if gain < -8388.608 or gain > 8388.607:
        raise ValueError
    if offset not in range(-128, 128):
        raise ValueError


class MapEntry:
    """Describe a openinverter parameter to CAN message mapping"""

    def __init__(
            self,
            param_id: int,
            position: int,
            length: int,
            gain: float,
            offset: int):

        _validate_map_entry_parameters(
            position,
            length,
            gain,
            offset
        )

        self.param_id = param_id
        self.position = position
        self.length = length
        self.gain = gain
        self.offset = offset

    def __eq__(self, other):
        if type(other) is type(self):
            return self.__dict__ == other.__dict__
        return False

    def __repr__(self) -> str:
        cls = self.__class__.__name__
        attrs = ", ".join(f"{k}={v}" for k, v in self.__dict__.items())
        return f"{cls}({attrs})"


def _validate_can_message_parameters(
        can_id: int,
        is_extended_frame: bool = False) -> None:
    if is_extended_frame:
        if can_id not in range(0, 0x20000000):
            raise ValueError
    else:
        if can_id not in range(0, 0x800):
            raise ValueError


class CanMessage:
    """
    A custom CAN message that maps openinverter parameters to a specific CAN ID
    """

    def __init__(
            self,
            can_id: int,
            params: List[MapEntry],
            is_extended_frame: bool = False) -> None:
        _validate_can_message_parameters(can_id, is_extended_frame)

        self.can_id = can_id
        self.params = params
        self.is_extended_frame = is_extended_frame

    def __eq__(self, other):
        if type(other) is type(self):
            return self.__dict__ == other.__dict__
        return False

    def __repr__(self) -> str:
        cls = self.__class__.__name__
        attrs = ", ".join(f"{k}={v}" for k, v in self.__dict__.items())
        return f"{cls}({attrs})"


class OpenInverterNode(BaseNode):
    """
    openinverter lightly abuses the CANopen SDO protocol to implement a series
    of command and control API end-points to manipulate and manage devices.
    This class wraps the raw protocol in a python API that masks the underlying
    complexities.
    """

    def __init__(self,
                 network: canopen.Network,
                 node_id: int,
                 object_dictionary: canopen.ObjectDictionary =
                 canopen.ObjectDictionary()
                 ) -> None:
        """Create temporary SDO client and attach to the network """

        super().__init__(node_id, object_dictionary)
        self.network = network
        self.node_id = node_id
        self.sdo = SdoClient(0x600 + node_id,
                             0x580 + node_id,
                             object_dictionary)
        self.sdo.network = network  # type: ignore
        network.subscribe(0x580 + node_id, self.sdo.on_response)

    def __del__(self) -> None:
        self.network.unsubscribe(0x580 + self.node_id)

    def serial_no(self) -> bytes:
        """Device unique serial number"""

        # Fetch the serial number in 3 parts reversing from little-endian on
        # the wire into a reversed array where the LSB is first and MSB is
        # last. This is odd but mirrors the behaviour of the STM32 terminal
        # "serial" command for consistency.
        serialno = bytearray()
        for i in reversed(range(3)):
            serialno.extend(
                reversed(self.sdo.upload(oi.SERIALNO_INDEX, i)))

        return serialno

    def save(self) -> None:
        """
        Request the remote node save its parameters to persistent storage
        """
        self.sdo.download(
            oi.COMMAND_INDEX, oi.SAVE_COMMAND_SUBINDEX, bytes(4))

    def load(self) -> None:
        """
        Request the remote node load its parameters from persistent storage
        """
        self.sdo.download(
            oi.COMMAND_INDEX, oi.LOAD_COMMAND_SUBINDEX, bytes(4))

    def reset(self) -> None:
        """Request the remote node reset/reboot"""
        self.sdo.download(
            oi.COMMAND_INDEX, oi.RESET_COMMAND_SUBINDEX, bytes(4))

    def load_defaults(self) -> None:
        """Request the remote node load its default parameters. Equivalent to
        a factory configuration reset."""
        self.sdo.download(
            oi.COMMAND_INDEX, oi.DEFAULTS_COMMAND_SUBINDEX, bytes(4))

    def start(self, mode: int = oi.START_MODE_NORMAL) -> None:
        """
        Request the remote node start normal operation

        :param mode: Device specific mode parameter
        """
        self.sdo.download(
            oi.COMMAND_INDEX,
            oi.START_COMMAND_SUBINDEX,
            UNSIGNED32.pack(mode))

    def stop(self) -> None:
        """Request the remote node stop normal operation"""
        self.sdo.download(
            oi.COMMAND_INDEX, oi.STOP_COMMAND_SUBINDEX, bytes(4))

    def add_can_map_entry(
            self,
            can_id: int,
            direction: Direction,
            param_id: int,
            position: int,
            length: int,
            gain: float,
            offset: int,
            is_extended_frame: bool = False) -> None:
        """
        Add a CAN map entry to transmit or receive the current value of a
        given parameter.

        :param can_id:    The CAN ID that will be transmitted.
                          [0,0x7ff] for standard frames
                          [0,0x1fffffff] for extended frames
        :param direction: The direction the parameter will be mapped, either
                          transmit or receive.
        :param param_id:  The openinverter parameter id to be mapped.
        :param position:  The bit position the parameter will start at. [0, 63]
        :param length:    The bit length the parameter will occupy. Positive
                          lengths indicate a little-endian length and negative
                          big-endian. [-32,-1][1, 32]
        :param gain:      The parameter is multiplied by the gain before being
                          inserted into the CAN frame. [-8388.608, 8388.607]
        :param offset:    The offset to be added to the parameter after the
                          gain is applied. [-128, 127]
        :param is_extended_frame: Does the can_id represent an extended CAN
                          frame id
        """
        _validate_can_message_parameters(can_id, is_extended_frame)
        _validate_map_entry_parameters(
            position,
            length,
            gain,
            offset
        )

        map_direction_index = {
            Direction.TX: oi.CAN_MAP_TX_INDEX,
            Direction.RX: oi.CAN_MAP_RX_INDEX
        }

        if direction in map_direction_index:
            cmd_index = map_direction_index[direction]
        else:
            raise ValueError

        # Fill out the SDO "variable" with the CAN ID we want to map
        if is_extended_frame:
            packed_can_id = can_id | oi.MAP_EXTENDED_FRAME_FLAG
        else:
            packed_can_id = can_id
        self.sdo.download(
            cmd_index,
            oi.MAP_CAN_ID_SUBINDEX,
            UNSIGNED32.pack(packed_can_id))

        # Fill out the SDO "variable" with the parameter ID to map and
        # the position and length they should take up in each CAN frame
        self.sdo.download(
            cmd_index,
            oi.MAP_PARAM_POS_LEN_SUBINDEX,
            struct.pack(
                "<HBb",
                param_id,
                position,
                length))

        # Finally fill out the SDO "variable" with the gain and offset
        # the parameter requires for the CAN frame. This will actually
        # cause the mapping to be created on the remote node
        gain_bytes = struct.pack("<i", int(gain * 1000))[:3]
        offset_bytes = struct.pack("<b", offset)
        self.sdo.download(
            cmd_index,
            oi.MAP_GAIN_OFFSET_SUBINDEX,
            gain_bytes + offset_bytes)

    def add_can_map(
            self,
            direction: Direction,
            msg_map: List[CanMessage]) -> None:
        """
        Add complete CAN map for a given direction. The map may be obtained
        from the list_can_map() method or manually constructed.

        :param direction: The direction the parameter will be mapped, either
                          transmit or receive.
        :param msg_map:  The list of CanMessage objects that comprise the map
        """
        for msg in msg_map:
            for param in msg.params:
                self.add_can_map_entry(
                    msg.can_id,
                    direction,
                    param.param_id,
                    param.position,
                    param.length,
                    param.gain,
                    param.offset,
                    msg.is_extended_frame
                )

    def _get_mapped_can_id(self, index: int) -> Optional[int]:
        """
        Get the can_id stored at a given index in a can message param map.

        :param index: Absolute CAN SDO index for the entry

        :return: The can_id at this map position or None if not present
        """
        try:
            can_id, = UNSIGNED32.unpack(
                self.sdo.upload(index, 0))
        except canopen.SdoAbortedError as err:
            if err.code == oi.SDO_ABORT_OBJECT_NOT_AVAILABLE:
                return None
            else:
                raise err

        return can_id

    def _get_map_entry(
            self,
            can_id_index: int,
            param_index: int) -> Optional[MapEntry]:
        """Retrieve the details of a specific parameter map entry"""
        # parameter mappings always occur in sequential pairs
        assert param_index % 2 == 1

        try:
            (param_id, position, length) = struct.unpack(
                "<HBb",
                self.sdo.upload(can_id_index, param_index))

            gain_offset_bytes = self.sdo.upload(
                can_id_index, param_index+1)

            # Sign-extend the 24-bit gain into a 32-bit signed integer
            gain_bytes = gain_offset_bytes[:3]
            neg_gain = (gain_bytes[2] & 0x80) > 0
            (gain,) = struct.unpack(
                "<i",
                gain_bytes + (b'\xff' if neg_gain else b'\x00'))

            # Scale fixed-point to a float
            gain = gain / 1000.0

            offset_bytes = gain_offset_bytes[3:4]
            (offset,) = struct.unpack("<b", offset_bytes)

            param = MapEntry(
                param_id,
                position,
                length,
                gain,
                offset)

        except canopen.SdoAbortedError as err:
            if err.code == oi.SDO_ABORT_OBJECT_NOT_AVAILABLE:
                return None
            else:
                raise err

        return param

    def _get_map_entries(
            self,
            can_id_index: int) -> List[MapEntry]:
        """List all of the parameter map entries for a specific CAN ID

        :param can_id_index: Absolute CAN SDO index for the message"""
        params = []

        param_index = 1
        while True:
            param = self._get_map_entry(can_id_index, param_index)

            if param is not None:
                params.append(param)
                param_index += 2
            else:
                break

        return params

    def list_can_map(self, direction: Direction) -> List[CanMessage]:
        """
        List all of the parameter to CAN frame mappings on the remote device.

        :param direction: The direction the map corresponds with.

        :return: A list of parameter to CAN message mappings.
        """

        if direction in MapListDirectionIndex:
            can_id_index = MapListDirectionIndex[direction]
        else:
            raise ValueError

        messages: List[CanMessage] = []
        while True:
            can_id = self._get_mapped_can_id(can_id_index)

            if can_id is not None:
                is_extended_frame = can_id & oi.MAP_EXTENDED_FRAME_FLAG > 0
                msg = CanMessage(
                    can_id & oi.MAP_EXTENDED_FRAME_MASK,
                    self._get_map_entries(can_id_index),
                    is_extended_frame)

                # Just ignore a message without any params
                if msg.params:
                    messages.append(msg)

                can_id_index += 1
            else:
                break

        return messages

    def remove_can_map_entry(
            self,
            direction: Direction,
            can_index: int,
            param_index: int) -> bool:
        """
        Remove a specific entry from a CAN parameter map.

        NOTE: This will invalidate the contents of a stored CAN map as the
        contents will be shuffled up automatically by the remote node.

        :param direction:   Which map direction to modify
        :param can_index:   The list index of the CAN ID to remove
        :param param_index: The list index of the param within the CanMessage
                            to remove

        :return: True if the removal was successful
        """

        if direction in MapListDirectionIndex:
            can_sdo_index = MapListDirectionIndex[direction]
        else:
            raise ValueError

        can_sdo_index += can_index

        # Removal is achieved by writing to the SDO index corresponding to the
        # param_id, position, offset parameter
        param_sdo_index = 2*(param_index+1)

        try:
            self.sdo.download(
                can_sdo_index, param_sdo_index, UNSIGNED32.pack(0))
        except canopen.SdoCommunicationError as err:
            if str(err) == "Unexpected response 0x23":
                return True
            else:
                raise err

        except canopen.SdoAbortedError as err:
            if err.code == oi.SDO_ABORT_OBJECT_NOT_AVAILABLE:
                return False
            else:
                raise err

        return False

    def clear_map(self, direction: Direction) -> None:
        """
        Remove all entries from the CAN message map

        :param direction:   Which map direction to clear
        """

        # Repeated removal of the 0th parameter of the 0th message will remove
        # everything as the device automatically shunts up existing parameters
        # into that position.
        while self.remove_can_map_entry(direction, 0, 0):
            _ = 1



"""Test cases relating to CAN upgrade of devices"""

import csv
import unittest
from pathlib import Path
from typing import List, Tuple
from unittest.mock import Mock, call

import canopen
import pytest

from openinverter_can_tool.can_upgrade import (CanUpgrader, Failure, State,
                                               StateUpdate)

TOOL = 0x7DD
DEVICE = 0x7DE

UPGRADE_DATA_DIR = Path(__file__).parent / "test_data" / "upgrade"

EMPTY_FIRMWARE = UPGRADE_DATA_DIR / "empty-firmware.bin"
MINIMAL_FIRMWARE = UPGRADE_DATA_DIR / "minimal-firmware.bin"
ONE_PAGE_FIRMWARE = UPGRADE_DATA_DIR / "one-page-firmware.bin"
TWO_PAGE_FIRMWARE = UPGRADE_DATA_DIR / "two-page-firmware.bin"

# Reduce test verbosity
# pylint: disable=missing-function-docstring


class TestFirmwareLoading:
    """Firmware loading specific tests that don't need a CAN network"""

    def test_non_existent_firmware_file_fails_to_initialise(self):
        with pytest.raises(FileNotFoundError):
            CanUpgrader(
                canopen.Network(),
                None,
                UPGRADE_DATA_DIR / "non-existent-file.bin")

    def test_sub_page_firmware_reports_page_count_of_one(self, tmp_path: Path):
        firmware = tmp_path / "firmware.bin"
        with open(firmware, "wb") as f:
            f.write(bytes(1023))

        upgrader = CanUpgrader(canopen.Network(), None, firmware)

        assert len(upgrader.pages) == 1

    def test_exact_page_size_firmware_reports_integer_page(
            self, tmp_path: Path):
        firmware = tmp_path / "firmware.bin"
        with open(firmware, "wb") as f:
            f.write(bytes(4096))

        upgrader = CanUpgrader(canopen.Network(), None, firmware)

        assert len(upgrader.pages) == 4

    def test_max_size_firmware_successfully_loads(
            self, tmp_path: Path):
        firmware = tmp_path / "firmware.bin"
        with open(firmware, "wb") as f:
            f.write(bytes(255*1024))

        upgrader = CanUpgrader(canopen.Network(), None, firmware)

        assert len(upgrader.pages) == 255

    def test_over_max_size_firmware_fails_to_load(
            self, tmp_path: Path):
        firmware = tmp_path / "firmware.bin"
        with open(firmware, "wb") as f:
            f.write(bytes(255*1024+1))

        with pytest.raises(ValueError, match="Firmware image too large"):
            CanUpgrader(canopen.Network(), None, firmware)

    def test_firmware_contents_are_padded_to_page_size_with_zeros(self):
        upgrader = CanUpgrader(canopen.Network(), None, MINIMAL_FIRMWARE)

        assert len(upgrader.pages) == 1
        assert upgrader.pages[0].data[0] == 0xa5
        assert upgrader.pages[0].data[1] == 0
        assert upgrader.pages[0].data[1023] == 0

    def test_firmware_with_exact_size_has_no_padding(self):
        upgrader = CanUpgrader(canopen.Network(), None, TWO_PAGE_FIRMWARE)

        assert len(upgrader.pages) == 2
        assert upgrader.pages[0].data[0] == 0x55
        assert upgrader.pages[1].data[1023] == 0xaa

    def test_minimal_firmware_crc_is_the_same_as_stm32_bootloader(self):
        upgrader = CanUpgrader(canopen.Network(), None, MINIMAL_FIRMWARE)

        # From a CAN capture of can-upgrader.py loading this firmware to
        # a real device
        assert upgrader.pages[0].crc == 0xF1B78CE3

    def test_one_page_firmware_crc_is_the_same_as_stm32_bootloader(self):
        upgrader = CanUpgrader(canopen.Network(), None, ONE_PAGE_FIRMWARE)

        # From a CAN capture of can-upgrader.py loading this firmware to
        # a real device
        assert upgrader.pages[0].crc == 0x8ADA4578

    def test_two_page_firmware_crc_is_the_same_as_stm32_bootloader(self):
        upgrader = CanUpgrader(canopen.Network(), None, TWO_PAGE_FIRMWARE)

        # From a CAN capture of can-upgrader.py loading this firmware to
        # a real device
        assert upgrader.pages[0].crc == 0x41B309C3
        assert upgrader.pages[1].crc == 0x1AB9F829


QUICK_TIMEOUT = 0.01
"""A really short timeout so the upgrade run finishes quickly"""


class TestCANUpgrade(unittest.TestCase):
    """
    Test the CAN upgrade state machine by example
    """

    def _send_message(self, can_id, data, remote=False):
        """Will be used instead of the usual Network.send_message method.

        Checks that the message data is as expected.
        """
        next_data = self.data.pop(0)
        self.assertEqual(next_data[0], TOOL, "No transmission was expected")
        self.assertSequenceEqual(data, next_data[1])
        self.assertEqual(can_id, TOOL)

        # pretend to use remote
        _ = remote

    def setUp(self):
        network = canopen.Network()
        network.send_message = self._send_message
        self.network = network
        self.data: List[Tuple[int, bytes]] = []

    def tearDown(self) -> None:
        # At the end of every test all of the data data should have been
        # consumed by _send_message()
        assert len(self.data) == 0

    def send_device_frames(self) -> None:
        """Send the CAN frames from the device(s). Replies are processed
        through the _send_message() hook"""
        while self.data and self.data[0][0] == DEVICE:
            self.network.notify(DEVICE, bytearray(self.data.pop(0)[1]), 0.0)

    def load_capture(self, capture: str) -> None:
        """Load a CAN capture from SavvyCAN in CSV format into the data"""
        with open(UPGRADE_DATA_DIR / capture,
                  newline="", encoding="utf-8") as capture_file:
            reader = csv.DictReader(capture_file)

            for row in reader:
                frame_id = int(row["ID"], 16)

                frame = ""
                for i in range(int(row["LEN"])):
                    frame += row[f"D{i+1}"]

                self.data.append((frame_id, bytes.fromhex(frame)))

    def test_unknown_device_is_ignored(self):
        self.data = [
            (DEVICE, b'\x33\x00\x00\x00\x29\x30\x19\x87')
        ]

        upgrader = CanUpgrader(self.network,
                               b"\x12\x34\x56\x78",
                               EMPTY_FIRMWARE)

        self.send_device_frames()

        assert not upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.START
        assert upgrader.serialno is None

    def test_recognise_specific_device_and_start_upgrade_process(self):
        self.data = [
            (DEVICE, b'\x33\x00\x00\x00\x29\x30\x19\x87'),
            (TOOL, b'\x29\x30\x19\x87')
        ]

        upgrader = CanUpgrader(self.network,
                               b"\x87\x19\x30\x29",
                               EMPTY_FIRMWARE)

        self.send_device_frames()

        assert not upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.HEADER
        assert upgrader.serialno == b"\x87\x19\x30\x29"

    def test_ignore_reserved_bytes_and_start_upgrade_process(self):
        self.data = [
            # Note: bytes 1, 2 and 3 have data in them rather than zero
            (DEVICE, b'\x33\x12\x34\x56\x29\x30\x19\x87'),
            (TOOL, b'\x29\x30\x19\x87')
        ]

        upgrader = CanUpgrader(self.network,
                               b"\x87\x19\x30\x29",
                               EMPTY_FIRMWARE)

        self.send_device_frames()

        assert not upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.HEADER
        assert upgrader.serialno == b"\x87\x19\x30\x29"

    def test_start_to_recover_the_next_device_to_boot(self):
        self.data = [
            (DEVICE, b'\x33\x00\x00\x00\x29\x30\x19\x87'),
            (TOOL, b'\x29\x30\x19\x87')
        ]

        upgrader = CanUpgrader(self.network, None, EMPTY_FIRMWARE)

        self.send_device_frames()

        assert not upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.HEADER
        assert upgrader.serialno == b"\x87\x19\x30\x29"

    def test_recognise_device_in_herd_and_start_upgrade_process(self):
        self.data = [
            (DEVICE, b'\x33\x00\x00\x00\x78\x56\x34\x12'),
            (DEVICE, b'\x33\x00\x00\x00\x29\x30\x19\x87'),
            (TOOL, b'\x29\x30\x19\x87'),
            (DEVICE, b'\x33\x00\x00\x00\x12\x34\x56\x78')
        ]
        upgrader = CanUpgrader(self.network,
                               b"\x87\x19\x30\x29",
                               EMPTY_FIRMWARE)

        self.send_device_frames()

        assert not upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.HEADER
        assert upgrader.serialno == b"\x87\x19\x30\x29"

    def test_zero_length_data_in_hello_frame_triggers_error_state(self):
        self.data = [(DEVICE, b'')]

        upgrader = CanUpgrader(self.network, None, EMPTY_FIRMWARE)

        self.send_device_frames()

        assert upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.FAILURE
        assert upgrader.failure == Failure.PROTOCOL_ERROR

    def test_random_data_in_hello_frame_triggers_error_state(self):
        self.data = [(DEVICE, b'\x12\x34\x56')]

        upgrader = CanUpgrader(self.network, None, EMPTY_FIRMWARE)

        self.send_device_frames()

        assert upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.FAILURE
        assert upgrader.failure == Failure.PROTOCOL_ERROR

    def test_unusually_short_hello_header_triggers_error_state(self):
        self.data = [(DEVICE, b'\x33')]

        upgrader = CanUpgrader(self.network, None, EMPTY_FIRMWARE)

        self.send_device_frames()

        assert upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.FAILURE
        assert upgrader.failure == Failure.PROTOCOL_ERROR

    def test_start_frame_indicates_upgrade_in_progress_error(self):
        self.data = [(DEVICE, b'\x53')]

        upgrader = CanUpgrader(self.network, None, EMPTY_FIRMWARE)

        self.send_device_frames()

        assert upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.FAILURE
        assert upgrader.failure == Failure.UPGRADE_IN_PROGRESS

    def test_page_frame_indicates_upgrade_in_progress_error(self):
        self.data = [(DEVICE, b'\x50')]

        upgrader = CanUpgrader(self.network, None, EMPTY_FIRMWARE)

        self.send_device_frames()

        assert upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.FAILURE
        assert upgrader.failure == Failure.UPGRADE_IN_PROGRESS

    def test_checksum_frame_indicates_upgrade_in_progress_error(self):
        self.data = [(DEVICE, b'\x43')]

        upgrader = CanUpgrader(self.network, None, EMPTY_FIRMWARE)

        self.send_device_frames()

        assert upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.FAILURE
        assert upgrader.failure == Failure.UPGRADE_IN_PROGRESS

    def test_done_frame_indicates_upgrade_in_progress_error(self):
        self.data = [(DEVICE, b'\x44')]

        upgrader = CanUpgrader(self.network, None, EMPTY_FIRMWARE)

        self.send_device_frames()

        assert upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.FAILURE
        assert upgrader.failure == Failure.UPGRADE_IN_PROGRESS

    def test_specific_device_reports_that_it_has_started(self):
        self.data = [
            (DEVICE, b'\x33\x00\x00\x00\x29\x30\x19\x87'),
            (TOOL, b'\x29\x30\x19\x87'),
            (DEVICE, b'S'),
            (TOOL, b'\x01')
        ]

        upgrader = CanUpgrader(self.network, None, MINIMAL_FIRMWARE)

        self.send_device_frames()

        assert not upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.UPLOAD

    def test_another_device_starting_during_start_request_is_ignored(self):
        self.data = [
            (DEVICE, b'\x33\x00\x00\x00\x29\x30\x19\x87'),
            (TOOL, b'\x29\x30\x19\x87'),
            (DEVICE, b'\x33\x00\x00\x00\x12\x34\x56\x78'),  # Extra HELLO
            (DEVICE, b'S'),
            (TOOL, b'\x01')
        ]

        upgrader = CanUpgrader(self.network, None, MINIMAL_FIRMWARE)

        self.send_device_frames()

        assert not upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.UPLOAD

    def test_device_send_empty_frame_instead_of_size_request(self):
        self.data = [
            (DEVICE, b'\x33\x00\x00\x00\x29\x30\x19\x87'),
            (TOOL, b'\x29\x30\x19\x87'),
            (DEVICE, b'')  # This is an invalid response from the device
        ]

        upgrader = CanUpgrader(self.network, None, MINIMAL_FIRMWARE)

        self.send_device_frames()

        assert upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.FAILURE
        assert upgrader.failure == Failure.PROTOCOL_ERROR

    def test_device_receives_first_8_bytes_of_firmware(self):
        self.data = [
            (DEVICE, b'\x33\x00\x00\x00\x29\x30\x19\x87'),
            (TOOL, b'\x29\x30\x19\x87'),
            (DEVICE, b'S'),
            (TOOL, b'\x01'),
            (DEVICE, b'P'),
            (TOOL, b'\xa5\x00\x00\x00\x00\x00\x00\x00')
        ]

        upgrader = CanUpgrader(self.network, None, MINIMAL_FIRMWARE)

        self.send_device_frames()

        assert not upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.UPLOAD

    def test_another_device_starting_during_page_request_is_ignored(self):
        self.data = [
            (DEVICE, b'\x33\x00\x00\x00\x29\x30\x19\x87'),
            (TOOL, b'\x29\x30\x19\x87'),
            (DEVICE, b'S'),
            (TOOL, b'\x01'),
            (DEVICE, b'\x33\x00\x00\x00\x12\x34\x56\x78'),  # Extra HELLO
            (DEVICE, b'P'),
            (TOOL, b'\xa5\x00\x00\x00\x00\x00\x00\x00')
        ]

        upgrader = CanUpgrader(self.network, None, MINIMAL_FIRMWARE)

        self.send_device_frames()

        assert not upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.UPLOAD

    def test_device_sends_empty_frame_instead_of_page_request(self):
        self.data = [
            (DEVICE, b'\x33\x00\x00\x00\x29\x30\x19\x87'),
            (TOOL, b'\x29\x30\x19\x87'),
            (DEVICE, b'S'),
            (TOOL, b'\x01'),
            (DEVICE, b'')  # This is an invalid response from the device
        ]

        upgrader = CanUpgrader(self.network, None, MINIMAL_FIRMWARE)

        self.send_device_frames()

        assert upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.FAILURE
        assert upgrader.failure == Failure.PROTOCOL_ERROR

    def test_device_sends_extra_data_in_page_request(self):
        self.data = [
            (DEVICE, b'\x33\x00\x00\x00\x29\x30\x19\x87'),
            (TOOL, b'\x29\x30\x19\x87'),
            (DEVICE, b'S'),
            (TOOL, b'\x01'),
            # This is unexpected so probably indicates a new incompatible
            # protocol we can't be expected to deal with
            (DEVICE, b'P\xaa')
        ]

        upgrader = CanUpgrader(self.network, None, MINIMAL_FIRMWARE)

        self.send_device_frames()

        assert upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.FAILURE
        assert upgrader.failure == Failure.PROTOCOL_ERROR

    def test_device_receives_complete_page_of_firmware(self):
        self.load_capture("minimal-firmware-upload-page.csv")

        upgrader = CanUpgrader(self.network, None, MINIMAL_FIRMWARE)

        self.send_device_frames()

        assert not upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.CHECK_CRC
        assert upgrader.progress == 0.0

    def test_device_receives_an_extra_page_data_request_which_fails(self):
        self.load_capture("minimal-firmware-over-request-page-data.csv")

        upgrader = CanUpgrader(self.network, None, MINIMAL_FIRMWARE)

        self.send_device_frames()

        assert upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.FAILURE
        assert upgrader.failure == Failure.PROTOCOL_ERROR
        assert upgrader.progress == 0.0

    def test_device_receives_crc_request_earlier_than_expected(self):
        self.load_capture("minimal-firmware-early-crc-request.csv")

        upgrader = CanUpgrader(self.network, None, MINIMAL_FIRMWARE)

        self.send_device_frames()

        assert upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.FAILURE
        assert upgrader.failure == Failure.PROTOCOL_ERROR
        assert upgrader.progress == 0.0

    def test_device_receives_page_of_firmware_but_fails_crc_check(self):
        self.load_capture("minimal-firmware-invalid-page-crc.csv")

        upgrader = CanUpgrader(self.network, None, MINIMAL_FIRMWARE)

        self.send_device_frames()

        assert upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.FAILURE
        assert upgrader.failure == Failure.PAGE_CRC_ERROR
        assert upgrader.progress == 0.0

    def test_device_completes_minimal_firmware_upgrade(self):
        self.load_capture("minimal-firmware.csv")

        upgrader = CanUpgrader(self.network, None, MINIMAL_FIRMWARE)

        self.send_device_frames()

        assert upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.COMPLETE
        assert upgrader.progress == 100.0

    def test_another_device_starting_during_checksum_request_is_ignored(self):
        self.load_capture("minimal-firmware-crc-extra-hello.csv")

        upgrader = CanUpgrader(self.network, None, MINIMAL_FIRMWARE)

        self.send_device_frames()

        assert upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.COMPLETE
        assert upgrader.progress == 100.0

    def test_another_device_starting_during_done_request_is_ignored(self):
        self.load_capture("minimal-firmware-done-extra-hello.csv")

        upgrader = CanUpgrader(self.network, None, MINIMAL_FIRMWARE)

        self.send_device_frames()

        assert upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.COMPLETE
        assert upgrader.progress == 100.0

    def test_successful_upgrade_of_device_with_one_page_firmware(self):
        self.load_capture("successful-one-page-upgrade.csv")

        upgrader = CanUpgrader(self.network, None, ONE_PAGE_FIRMWARE)

        self.send_device_frames()

        assert upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.COMPLETE
        assert upgrader.progress == 100.0

    def test_successful_upgrade_of_device_with_two_page_firmware(self):
        self.load_capture("successful-two-page-upgrade.csv")

        upgrader = CanUpgrader(self.network, None, TWO_PAGE_FIRMWARE)

        self.send_device_frames()

        assert upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.COMPLETE
        assert upgrader.progress == 100.0

    def test_zero_length_firmware_succeeds_and_sends_no_pages(self):
        self.data = [
            (DEVICE, b'\x33\x00\x00\x00\x29\x30\x19\x87'),
            (TOOL, b'\x29\x30\x19\x87'),
            (DEVICE, b'S'),
            (TOOL, b'\x00'),
            (DEVICE, b'D'),
        ]

        upgrader = CanUpgrader(self.network, None, EMPTY_FIRMWARE)

        self.send_device_frames()

        assert upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.COMPLETE
        assert upgrader.progress == 100.0

    def test_state_transition_notifications_are_sent_during_an_upgrade(self):
        self.load_capture("successful-one-page-upgrade.csv")
        callback = Mock()

        upgrader = CanUpgrader(self.network, None, ONE_PAGE_FIRMWARE, callback)

        self.send_device_frames()

        assert upgrader.run(QUICK_TIMEOUT)
        serialno = b"\x87\x19\x30\x29"
        callback_list = [
            call(StateUpdate(State.START, None, None, 0.0)),
            call(StateUpdate(State.HEADER, serialno, None, 0.0)),
            call(StateUpdate(State.UPLOAD, serialno, None, 0.0)),
            call(StateUpdate(State.CHECK_CRC, serialno, None, 0.0)),
            call(StateUpdate(State.WAIT_FOR_DONE, serialno, None, 0.0)),
            call(StateUpdate(State.COMPLETE, serialno, None, 100.0))
        ]
        callback.assert_has_calls(callback_list)

    def test_failure_notification_contains_failure_code(self):
        self.data = [(DEVICE, b'\x43')]
        callback = Mock()

        upgrader = CanUpgrader(self.network, None, EMPTY_FIRMWARE, callback)

        self.send_device_frames()

        assert upgrader.run(QUICK_TIMEOUT)
        callback.assert_called_with(
            StateUpdate(State.FAILURE, None, Failure.UPGRADE_IN_PROGRESS, 0.0))

    def test_progress_of_during_upgrade_with_two_page_firmware(self):
        self.load_capture("successful-two-page-upgrade.csv")

        callback = Mock()

        upgrader = CanUpgrader(self.network, None, TWO_PAGE_FIRMWARE, callback)

        self.send_device_frames()

        assert upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.COMPLETE
        assert upgrader.progress == 100.0

        serialno = b"\x87\x19\x30\x29"
        callback_list = [
            call(StateUpdate(State.START, None, None, 0.0)),
            call(StateUpdate(State.HEADER, serialno, None, 0.0)),
            call(StateUpdate(State.UPLOAD, serialno, None, 0.0)),
            call(StateUpdate(State.CHECK_CRC, serialno, None, 0.0)),
            call(StateUpdate(State.UPLOAD, serialno, None, 0.0)),
            call(StateUpdate(State.CHECK_CRC, serialno, None, 50.0)),
            call(StateUpdate(State.WAIT_FOR_DONE, serialno, None, 50.0)),
            call(StateUpdate(State.COMPLETE, serialno, None, 100.0))
        ]
        callback.assert_has_calls(callback_list)

    def test_extra_data_sent_after_completion_is_ignored(self):
        self.data = [
            (DEVICE, b'\x33\x00\x00\x00\x29\x30\x19\x87'),
            (TOOL, b'\x29\x30\x19\x87'),
            (DEVICE, b'S'),
            (TOOL, b'\x00'),
            (DEVICE, b'D'),
            (DEVICE, b'\x33\x00\x00\x00\x29\x30\x19\x87')
        ]

        upgrader = CanUpgrader(self.network, None, EMPTY_FIRMWARE)

        self.send_device_frames()

        assert upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.COMPLETE
        assert upgrader.progress == 100.0

    def test_extra_data_after_failure_is_ignored(self):
        self.data = [
            (DEVICE, b'\x12\x34\x56'),
            (DEVICE, b'S')
        ]

        upgrader = CanUpgrader(self.network, None, EMPTY_FIRMWARE)

        self.send_device_frames()

        assert upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.FAILURE
        assert upgrader.failure == Failure.PROTOCOL_ERROR

    def test_device_sending_invalid_response_instead_of_done_fails(self):
        self.data = [
            (DEVICE, b'\x33\x00\x00\x00\x29\x30\x19\x87'),
            (TOOL, b'\x29\x30\x19\x87'),
            (DEVICE, b'S'),
            (TOOL, b'\x00'),
            (DEVICE, b'')  # This is an invalid response from the device
        ]

        upgrader = CanUpgrader(self.network, None, EMPTY_FIRMWARE)

        self.send_device_frames()

        assert upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.FAILURE
        assert upgrader.failure == Failure.PROTOCOL_ERROR

    def test_two_page_firmware_fails_with_crc_error_on_first_page(self):
        self.load_capture("two-page-upgrade-fails-crc-after-first-page.csv")

        upgrader = CanUpgrader(self.network, None, TWO_PAGE_FIRMWARE)

        self.send_device_frames()

        assert upgrader.run(QUICK_TIMEOUT)
        assert upgrader.state == State.FAILURE
        assert upgrader.failure == Failure.PAGE_CRC_ERROR
        assert upgrader.progress == 0.0
